"""Wrapper classes for the chemprop components to make them compatible with scikit-learn."""

import abc
from typing import Any, Iterable, Self

import torch
from chemprop.conf import DEFAULT_ATOM_FDIM, DEFAULT_BOND_FDIM, DEFAULT_HIDDEN_DIM
from chemprop.models.model import MPNN as _MPNN
from chemprop.nn.agg import Aggregation
from chemprop.nn.agg import MeanAggregation as _MeanAggregation
from chemprop.nn.agg import SumAggregation as _SumAggregation
from chemprop.nn.ffn import MLP
from chemprop.nn.loss import LossFunction
from chemprop.nn.message_passing import BondMessagePassing as _BondMessagePassing
from chemprop.nn.message_passing import MessagePassing
from chemprop.nn.metrics import (
    BinaryAUROCMetric,
    CrossEntropyMetric,
    Metric,
    MSEMetric,
    SIDMetric,
)
from chemprop.nn.predictors import BinaryClassificationFFN as _BinaryClassificationFFN
from chemprop.nn.predictors import BinaryDirichletFFN as _BinaryDirichletFFN
from chemprop.nn.predictors import EvidentialFFN as _EvidentialFFN
from chemprop.nn.predictors import (
    MulticlassClassificationFFN as _MulticlassClassificationFFN,
)
from chemprop.nn.predictors import MulticlassDirichletFFN as _MulticlassDirichletFFN
from chemprop.nn.predictors import MveFFN as _MveFFN
from chemprop.nn.predictors import RegressionFFN as _RegressionFFN
from chemprop.nn.predictors import SpectralFFN as _SpectralFFN
from chemprop.nn.predictors import (
    _FFNPredictorBase as _Predictor,  # pylint: disable=protected-access
)
from chemprop.nn.transforms import UnscaleTransform
from chemprop.nn.utils import Activation, get_activation_function
from sklearn.base import BaseEstimator
from torch import Tensor, nn

from molpipeline.estimators.chemprop.loss_wrapper import (
    BCELoss,
    BinaryDirichletLoss,
    CrossEntropyLoss,
    EvidentialLoss,
    MSELoss,
    MulticlassDirichletLoss,
    MVELoss,
    SIDLoss,
)


# pylint: disable=too-many-ancestors, too-many-instance-attributes
class BondMessagePassing(_BondMessagePassing, BaseEstimator):
    """A wrapper for the BondMessagePassing class."""

    def __init__(
        self,
        d_v: int = DEFAULT_ATOM_FDIM,
        d_e: int = DEFAULT_BOND_FDIM,
        d_h: int = DEFAULT_HIDDEN_DIM,
        bias: bool = False,
        depth: int = 3,
        dropout_rate: float = 0.0,
        activation: str | Activation = Activation.RELU,
        undirected: bool = False,
        d_vd: int | None = None,
    ):
        """Initialize the BondMessagePassing class.

        Parameters
        ----------
        d_v : int, optional (default=DEFAULT_ATOM_FDIM)
            The input vertices feature dimension, by default DEFAULT_ATOM_FDIM
        d_e : int, optional (default=DEFAULT_BOND_FDIM)
            The input edges feature dimension, by default DEFAULT_BOND_FDIM
        d_h : int, optional (default=DEFAULT_HIDDEN_DIM)
            The hidden layer dimension, by default DEFAULT_HIDDEN_DIM
        bias : bool, optional (default=False)
            Whether to use bias in the weight matrices.
        depth : int, optional (default=3)
            Number of message passing layers.
        dropout_rate : float, optional (default=0)
            Dropout rate.
        activation : str or Activation, optional (default=Activation.RELU)
            Activation function.
        undirected : bool, optional (default=False)
            Whether to use undirected edges.
        d_vd : int or None, optional (default=None)
            Dimension of additional vertex descriptors that will be concatenated to the hidden features before readout
        """
        super().__init__(
            d_v,
            d_e,
            d_h,
            bias,
            depth,
            dropout_rate,
            activation,
            undirected,
            d_vd,
        )
        self.d_v = d_v
        self.d_e = d_e
        self.d_h = d_h
        self.d_vd = d_vd
        self.bias = bias
        self.activation = activation
        self.dropout_rate = dropout_rate

    def reinitialize_network(self) -> Self:
        """Reinitialize the network with the current parameters.

        Returns
        -------
        Self
            The reinitialized network.
        """
        self.W_i, self.W_h, self.W_o, self.W_d = self.setup(
            self.d_v, self.d_e, self.d_h, self.d_vd, self.bias
        )
        self.dropout = nn.Dropout(self.dropout_rate)
        if isinstance(self.activation, str):
            self.tau = get_activation_function(self.activation)
        else:
            self.tau = self.activation
        return self

    def set_params(self, **params: Any) -> Self:
        """Set the parameters of the class and update the network.

        Parameters
        ----------
        **params: Any
            The parameters to set.

        Returns
        -------
        Self
            The model with the new parameters.
        """
        super().set_params(**params)
        self.reinitialize_network()
        return self


# pylint: disable=too-many-ancestors, too-many-instance-attributes
class PredictorWrapper(_Predictor, BaseEstimator, abc.ABC):  # type: ignore
    """Abstract wrapper for the Predictor class."""

    _T_default_criterion: LossFunction
    _T_default_metric: Metric

    def __init__(  # pylint: disable=too-many-positional-arguments
        self,
        n_tasks: int = 1,
        input_dim: int = DEFAULT_HIDDEN_DIM,
        hidden_dim: int = 300,
        n_layers: int = 1,
        dropout: float = 0,
        activation: str = "relu",
        criterion: LossFunction | None = None,
        task_weights: Tensor | None = None,
        threshold: float | None = None,
        output_transform: UnscaleTransform | None = None,
        **kwargs: Any,
    ):
        """Initialize the BinaryClassificationFFN class.

        Parameters
        ----------
        n_tasks : int, optional (default=1)
            Number of tasks.
        input_dim : int, optional (default=DEFAULT_HIDDEN_DIM)
            Input dimension.
        hidden_dim : int, optional (default=300)
            Hidden dimension.
        n_layers : int, optional (default=1)
            Number of layers.
        dropout : float, optional (default=0)
            Dropout rate.
        activation : str, optional (default="relu")
            Activation function.
        criterion : LossFunction or None, optional (default=None)
            Loss function. None defaults to BCELoss.
        task_weights : Tensor or None, optional (default=None)
            Task weights.
        threshold : float or None, optional (default=None)
            Threshold for binary classification.
        output_transform : UnscaleTransform or None, optional (default=None)
            Transformations to apply to the output. None defaults to UnscaleTransform.
        kwargs : Any
            Additional keyword arguments.
        """
        if task_weights is None:
            task_weights = torch.ones(n_tasks)
        super().__init__(
            n_tasks=n_tasks,
            input_dim=input_dim,
            hidden_dim=hidden_dim,
            n_layers=n_layers,
            dropout=dropout,
            activation=activation,
            criterion=criterion,
            output_transform=output_transform,
            **kwargs,
        )
        self.n_tasks = n_tasks
        self._input_dim = input_dim
        self.hidden_dim = hidden_dim
        self.n_layers = n_layers
        self.dropout = dropout
        self.activation = activation
        self.task_weights = task_weights
        self.threshold = threshold

    @property
    def input_dim(self) -> int:
        """Get the dimension of input."""
        return self._input_dim

    @input_dim.setter
    def input_dim(self, value: int) -> None:
        """Set the dimension of input.

        Parameters
        ----------
        value : int
            The dimension of input.
        """
        self._input_dim = value

    @property
    def n_tasks(self) -> int:
        """Get the number of tasks."""
        return self._n_tasks

    @n_tasks.setter
    def n_tasks(self, value: int) -> None:
        """Set the number of tasks.

        Parameters
        ----------
        value : int
            The number of tasks.
        """
        self._n_tasks = value

    def reinitialize_fnn(self) -> Self:
        """Reinitialize the feedforward network.

        Returns
        -------
        Self
            The reinitialized feedforward network.
        """
        self.ffn = MLP.build(
            input_dim=self.input_dim,
            output_dim=self.n_tasks * self.n_targets,
            hidden_dim=self.hidden_dim,
            n_layers=self.n_layers,
            dropout=self.dropout,
            activation=self.activation,
        )
        return self

    def set_params(self, **params: Any) -> Self:
        """Set the parameters of the class and reinitialize the feedforward network.

        Parameters
        ----------
        **params: Any
            The parameters to set.

        Returns
        -------
        Self
            The model with the new parameters.
        """
        super().set_params(**params)
        self.reinitialize_fnn()
        return self


class RegressionFFN(PredictorWrapper, _RegressionFFN):  # type: ignore
    """A wrapper for the RegressionFFN class."""

    n_targets: int = 1
    _T_default_criterion = MSELoss
    _T_default_metric = MSEMetric


class MveFFN(PredictorWrapper, _MveFFN):  # type: ignore
    """A wrapper for the MveFFN class."""

    n_targets: int = 2
    _T_default_criterion = MVELoss


class EvidentialFFN(PredictorWrapper, _EvidentialFFN):  # type: ignore
    """A wrapper for the EvidentialFFN class."""

    n_targets: int = 4
    _T_default_criterion = EvidentialLoss


class BinaryClassificationFFN(PredictorWrapper, _BinaryClassificationFFN):  # type: ignore
    """A wrapper for the BinaryClassificationFFN class."""

    n_targets: int = 1
    _T_default_criterion = BCELoss
    _T_default_metric = BinaryAUROCMetric


class BinaryDirichletFFN(PredictorWrapper, _BinaryDirichletFFN):  # type: ignore
    """A wrapper for the BinaryDirichletFFN class."""

    n_targets: int = 2
    _T_default_criterion = BinaryDirichletLoss
    _T_default_metric = BinaryAUROCMetric


class MulticlassClassificationFFN(PredictorWrapper, _MulticlassClassificationFFN):  # type: ignore
    """A wrapper for the MulticlassClassificationFFN class."""

    n_targets: int = 1
    _T_default_criterion = CrossEntropyLoss
    _T_default_metric = CrossEntropyMetric

    def __init__(  # pylint: disable=too-many-positional-arguments
        self,
        n_classes: int,
        n_tasks: int = 1,
        input_dim: int = DEFAULT_HIDDEN_DIM,
        hidden_dim: int = 300,
        n_layers: int = 1,
        dropout: float = 0.0,
        activation: str = "relu",
        criterion: LossFunction | None = None,
        task_weights: Tensor | None = None,
        threshold: float | None = None,
        output_transform: UnscaleTransform | None = None,
    ):
        """Initialize the MulticlassClassificationFFN class.

        Parameters
        ----------
        n_classes : int
            how many classes are expected in the output
        n_tasks : int, optional (default=1)
            Number of tasks.
        input_dim : int, optional (default=DEFAULT_HIDDEN_DIM)
            Input dimension.
        hidden_dim : int, optional (default=300)
            Hidden dimension.
        n_layers : int, optional (default=1)
            Number of layers.
        dropout : float, optional (default=0)
            Dropout rate.
        activation : str, optional (default="relu")
            Activation function.
        criterion : LossFunction or None, optional (default=None)
            Loss function. None defaults to BCELoss.
        task_weights : Tensor or None, optional (default=None)
            Task weights.
        threshold : float or None, optional (default=None)
            Threshold for binary classification.
        output_transform : UnscaleTransform or None, optional (default=None)
            Transformations to apply to the output. None defaults to UnscaleTransform.
        """
        super().__init__(
            n_tasks,
            input_dim,
            hidden_dim,
            n_layers,
            dropout,
            activation,
            criterion,
            task_weights,
            threshold,
            output_transform,
            n_classes=n_classes,
        )

        self.n_classes = n_classes


class MulticlassDirichletFFN(PredictorWrapper, _MulticlassDirichletFFN):  # type: ignore
    """A wrapper for the MulticlassDirichletFFN class."""

    n_targets: int = 1
    _T_default_criterion = MulticlassDirichletLoss
    _T_default_metric = CrossEntropyMetric


class SpectralFFN(PredictorWrapper, _SpectralFFN):  # type: ignore
    """A wrapper for the SpectralFFN class."""

    n_targets: int = 1
    _T_default_criterion = SIDLoss
    _T_default_metric = SIDMetric


class MPNN(_MPNN, BaseEstimator):
    """A wrapper for the MPNN class.

    The MPNN is the main model class in chemprop. It consists of a message passing network, an aggregation function,
    and a feedforward network for prediction.
    """

    bn: nn.BatchNorm1d | nn.Identity

    def __init__(
        self,
        message_passing: MessagePassing,
        agg: Aggregation,
        predictor: PredictorWrapper,
        batch_norm: bool = True,
        metric_list: Iterable[Metric] | None = None,
        warmup_epochs: int = 2,
        init_lr: float = 1e-4,
        max_lr: float = 1e-3,
        final_lr: float = 1e-4,
    ):
        """Initialize the MPNN class.

        Parameters
        ----------
        message_passing : MessagePassing
            The message passing network.
        agg : Aggregation
            The aggregation function.
        predictor : Predictor
            The predictor function.
        batch_norm : bool, optional (default=True)
            Whether to use batch normalization.
        metric_list : Iterable[Metric] | None, optional (default=None)
            The metrics to use for evaluation.
        warmup_epochs : int, optional (default=2)
            The number of epochs to use for the learning rate warmup.
        init_lr : float, optional (default=1e-4)
            The initial learning rate.
        max_lr : float, optional (default=1e-3)
            The maximum learning rate.
        final_lr : float, optional (default=1e-4)
            The final learning rate.
        """
        super().__init__(
            message_passing=message_passing,
            agg=agg,
            predictor=predictor,
            batch_norm=batch_norm,
            metrics=metric_list,
            warmup_epochs=warmup_epochs,
            init_lr=init_lr,
            max_lr=max_lr,
            final_lr=final_lr,
        )
        self.metric_list = metric_list
        self.batch_norm = batch_norm

    def reinitialize_network(self) -> Self:
        """Reinitialize the network with the current parameters.

        Returns
        -------
        Self
            The reinitialized network.
        """
        if self.batch_norm:
            self.bn = nn.BatchNorm1d(self.message_passing.output_dim)
        else:
            self.bn = nn.Identity()

        if self.metric_list is None:
            # pylint: disable=protected-access
            self.metrics = [self.predictor._T_default_metric, self.criterion]
        else:
            self.metrics = list(self.metric_list) + [self.criterion]

        return self

    def set_params(self, **params: Any) -> Self:
        """Set the parameters of the class and update the network.

        Parameters
        ----------
        **params: Any
            The parameters to set.

        Returns
        -------
        Self
            The model with the new parameters.
        """
        super().set_params(**params)
        self.reinitialize_network()
        return self


# pylint: disable=too-many-ancestors
class MeanAggregation(_MeanAggregation, BaseEstimator):
    """Aggregate the graph-level representation by averaging the node representations."""

    def __init__(self, dim: int = 0):
        """Initialize the MeanAggregation class.

        Parameters
        ----------
        dim : int, optional (default=0)
            The dimension to aggregate over. See torch_scater.scatter for more details.
        """
        super().__init__(dim)


# pylint: disable=too-many-ancestors
class SumAggregation(_SumAggregation, BaseEstimator):
    """Aggregate the graph-level representation by summing the node representations."""

    def __init__(self, dim: int = 0):
        """Initialize the SumAggregation class.

        Parameters
        ----------
        dim : int, optional (default=0)
            The dimension to aggregate over. See torch_scater.scatter for more details.
        """
        super().__init__(dim)



"""Abstract classes for transforming rdkit molecules to bit vectors."""

from __future__ import annotations  # for all the python 3.8 users out there.

import abc
import copy
from typing import Any, Iterable, Literal, Optional, get_args, overload

try:
    from typing import Self, TypeAlias  # type: ignore[attr-defined]
except ImportError:
    from typing_extensions import Self, TypeAlias

import numpy as np
import numpy.typing as npt
from rdkit.Chem import rdFingerprintGenerator
from rdkit.DataStructs import ExplicitBitVect
from scipy import sparse

from molpipeline.abstract_pipeline_elements.core import MolToAnyPipelineElement
from molpipeline.utils.matrices import sparse_from_index_value_dicts
from molpipeline.utils.molpipeline_types import RDKitMol
from molpipeline.utils.substructure_handling import CircularAtomEnvironment

# possible output types for a fingerprint:
# - "sparse" is a sparse csr_matrix
# - "dense" is a numpy array
# - "explicit_bit_vect" is a list of RDKit's ExplicitBitVect
OutputDatatype: TypeAlias = Literal["sparse", "dense", "explicit_bit_vect"]


class MolToFingerprintPipelineElement(MolToAnyPipelineElement, abc.ABC):
    """Abstract class for PipelineElements which transform molecules to integer vectors."""

    _n_bits: int
    _feature_names: list[str]
    _output_type = "binary"
    _return_as: OutputDatatype

    def __init__(
        self,
        return_as: OutputDatatype = "sparse",
        name: str = "MolToFingerprintPipelineElement",
        n_jobs: int = 1,
        uuid: Optional[str] = None,
    ):
        """Initialize abstract class.

        Parameters
        ----------
        return_as: Literal["sparse", "dense", "explicit_bit_vect"]
            Type of output. When "sparse" the fingerprints will be returned as a scipy.sparse.csr_matrix
            holding a sparse representation of the bit vectors. With "dense" a numpy matrix will be returned.
            With "explicit_bit_vect" the fingerprints will be returned as a list of RDKit's
            rdkit.DataStructs.cDataStructs.ExplicitBitVect.
        name: str
            Name of PipelineElement.
        n_jobs:
            Number of jobs.
        uuid: Optional[str]
            Unique identifier.
        """
        super().__init__(
            name=name,
            n_jobs=n_jobs,
            uuid=uuid,
        )
        self._return_as = return_as

    @property
    def n_bits(self) -> int:
        """Get number of bits in (or size of) fingerprint."""
        return self._n_bits

    @property
    def feature_names(self) -> list[str]:
        """Get feature names."""
        return self._feature_names[:]

    @overload
    def assemble_output(  # type: ignore
        self, value_list: Iterable[npt.NDArray[np.int_]]
    ) -> npt.NDArray[np.int_]: ...

    @overload
    def assemble_output(
        self, value_list: Iterable[dict[int, int]]
    ) -> sparse.csr_matrix: ...

    @overload
    def assemble_output(
        self, value_list: Iterable[ExplicitBitVect]
    ) -> list[ExplicitBitVect]: ...

    def assemble_output(
        self,
        value_list: (
            Iterable[dict[int, int]]
            | Iterable[npt.NDArray[np.int_]]
            | Iterable[ExplicitBitVect]
        ),
    ) -> sparse.csr_matrix | npt.NDArray[np.int_] | list[ExplicitBitVect]:
        """Transform output of all transform_single operations to matrix.

        Parameters
        ----------
        value_list:  Iterable[dict[int, int]] | Iterable[npt.NDArray[np.int_]] | Iterable[ExplicitBitVect]
            Either Iterable of dicts which encode the rows of the feature matrix.
            Keys: column index, values: column value. Each dict represents one molecule.
            Or an Iterable of RDKit's ExplicitBitVect or an Iterable of numpy arrays representing the
            fingerprint list.

        Returns
        -------
        sparse.csr_matrix | npt.NDArray[np.int_] | list[ExplicitBitVect]
            Matrix of Morgan-fingerprint features.
        """
        if self._return_as == "explicit_bit_vect":
            # return as list of RDkit's ExplicitBitVect
            return list(value_list)  # type: ignore
        if self._return_as == "dense":
            # return dense numpy matrix
            return np.vstack(list(value_list))  # type: ignore

        # convert dict representation to csr_matrix
        return sparse_from_index_value_dicts(value_list, self._n_bits)  # type: ignore

    def get_params(self, deep: bool = True) -> dict[str, Any]:
        """Get object parameters relevant for copying the class.

        Parameters
        ----------
        deep: bool
            If True get a deep copy of the parameters.

        Returns
        -------
        dict[str, Any]
            Dictionary of parameter names and values.
        """
        parameters = super().get_params(deep)
        if deep:
            parameters["return_as"] = copy.copy(self._return_as)
        else:
            parameters["return_as"] = self._return_as

        return parameters

    def set_params(self, **parameters: Any) -> Self:
        """Set object parameters relevant for copying the class.

        Parameters
        ----------
        parameters: Any
            Dictionary of parameter names and values.

        Returns
        -------
        Self
            Copied object with updated parameters.
        """
        parameter_dict_copy = dict(parameters)
        return_as = parameter_dict_copy.pop("return_as", None)
        if return_as is not None:
            if return_as not in get_args(OutputDatatype):
                raise ValueError(
                    f"return_as has to be one of {get_args(OutputDatatype)}! (Received: {return_as})"
                )
            self._return_as = return_as
        super().set_params(**parameter_dict_copy)
        return self

    def transform(self, values: list[RDKitMol]) -> sparse.csr_matrix:
        """Transform the list of molecules to sparse matrix of Morgan-fingerprint features.

        Parameters
        ----------
        values: list[RDKitMol]
            List of RDKit molecules which are transformed to a sparse matrix.

        Returns
        -------
        sparse.csr_matrix
            Sparse matrix of Morgan-fingerprint features.
        """
        return super().transform(values)

    @abc.abstractmethod
    def pretransform_single(
        self, value: RDKitMol
    ) -> dict[int, int] | npt.NDArray[np.int_] | ExplicitBitVect:
        """Transform mol to dict, where items encode columns indices and values, respectively.

        Parameters
        ----------
        value: RDKitMol
            RDKit molecule which is encoded by the fingerprint.

        Returns
        -------
        dict[int, int]
            Dictionary to encode row in matrix. Keys: column index, values: column value.
        """


class MolToRDKitGenFPElement(MolToFingerprintPipelineElement, abc.ABC):
    """Abstract class for PipelineElements using the FingeprintGenerator64."""

    def __init__(
        self,
        counted: bool = False,
        return_as: OutputDatatype = "sparse",
        name: str = "MolToRDKitGenFin",
        n_jobs: int = 1,
        uuid: Optional[str] = None,
    ):
        """Initialize abstract class.

        Parameters
        ----------
        counted: bool
            Whether to count the bits or not.
        return_as: Literal["sparse", "dense", "explicit_bit_vect"]
            Type of output. When "sparse" the fingerprints will be returned as a scipy.sparse.csr_matrix
        name: str
            Name of PipelineElement.
        n_jobs:
            Number of jobs.
        uuid: Optional[str]
            Unique identifier.
        """
        super().__init__(
            return_as=return_as,
            name=name,
            n_jobs=n_jobs,
            uuid=uuid,
        )
        self.counted = counted

    @abc.abstractmethod
    def _get_fp_generator(self) -> rdFingerprintGenerator.FingerprintGenerator64:
        """Get fingerprint generator.

        Returns
        -------
        rdFingerprintGenerator.FingerprintGenerator64
            Fingerprint generator.
        """

    def pretransform_single(
        self, value: RDKitMol
    ) -> ExplicitBitVect | npt.NDArray[np.int_] | dict[int, int]:
        """Transform a single compound to a dictionary.

        Keys denote the feature position, values the count. Here always 1.

        Parameters
        ----------
        value: RDKitMol
            Molecule for which the fingerprint is generated.

        Returns
        -------
        ExplicitBitVect | npt.NDArray[np.int_] | dict[int, int]
            If return_as is "explicit_bit_vect" return ExplicitBitVect.
            If return_as is "dense" return numpy array.
            If return_as is "sparse" return dictionary with feature-position as key and count as value.
        """
        fingerprint_generator = self._get_fp_generator()
        if self._return_as == "dense":
            if self.counted:
                return fingerprint_generator.GetCountFingerprintAsNumPy(value)
            return fingerprint_generator.GetFingerprintAsNumPy(value)

        if self.counted:
            fingerprint = fingerprint_generator.GetCountFingerprint(value)
        else:
            fingerprint = fingerprint_generator.GetFingerprint(value)

        if self._return_as == "explicit_bit_vect":
            return fingerprint

        if self.counted:
            return fingerprint.GetNonzeroElements()

        return {pos: 1 for pos in fingerprint.GetOnBits()}

    def get_params(self, deep: bool = True) -> dict[str, Any]:
        """Get object parameters relevant for copying the class.

        Parameters
        ----------
        deep: bool
            If True get a deep copy of the parameters.

        Returns
        -------
        dict[str, Any]
            Dictionary of parameter names and values.
        """
        parameters = super().get_params(deep)
        if deep:
            parameters["counted"] = bool(self.counted)
        else:
            parameters["counted"] = self.counted

        return parameters

    def set_params(self, **parameters: Any) -> Self:
        """Set object parameters relevant for copying the class.

        Parameters
        ----------
        parameters: Any
            Dictionary of parameter names and values.

        Returns
        -------
        Self
            Copied object with updated parameters.
        """
        parameter_dict_copy = dict(parameters)
        counted = parameter_dict_copy.pop("counted", None)
        if counted is not None:
            self.counted = bool(counted)
        super().set_params(**parameter_dict_copy)
        return self


class ABCMorganFingerprintPipelineElement(MolToRDKitGenFPElement, abc.ABC):
    """Abstract Class for Morgan fingerprints."""

    @property
    def output_type(self) -> str:
        """Get output type."""
        if self.counted:
            return "integer"
        return "binary"

    # pylint: disable=R0913
    def __init__(
        self,
        radius: int = 2,
        use_features: bool = False,
        counted: bool = False,
        return_as: Literal["sparse", "dense", "explicit_bit_vect"] = "sparse",
        name: str = "AbstractMorgan",
        n_jobs: int = 1,
        uuid: Optional[str] = None,
    ):
        """Initialize abstract class.

        Parameters
        ----------
        radius: int
            Radius of fingerprint.
        use_features: bool
            Whether to represent atoms by element or category (donor, acceptor. etc.)
        counted: bool
            Whether to count the bits or not.
        return_as: Literal["sparse", "dense", "explicit_bit_vect"]
            Type of output. When "sparse" the fingerprints will be returned as a scipy.sparse.csr_matrix
            holding a sparse representation of the bit vectors. With "dense" a numpy matrix will be returned.
            With "explicit_bit_vect" the fingerprints will be returned as a list of RDKit's
            rdkit.DataStructs.cDataStructs.ExplicitBitVect.
        name: str
            Name of PipelineElement.
        n_jobs:
            Number of jobs.
        uuid: Optional[str]
            Unique identifier.
        """
        # pylint: disable=R0801
        super().__init__(
            return_as=return_as,
            counted=counted,
            name=name,
            n_jobs=n_jobs,
            uuid=uuid,
        )
        self._use_features = use_features
        if isinstance(radius, int) and radius >= 0:
            self._radius = radius
        else:
            raise ValueError(
                f"Number of bits has to be a positive integer! (Received: {radius})"
            )

    def get_params(self, deep: bool = True) -> dict[str, Any]:
        """Get object parameters relevant for copying the class.

        Parameters
        ----------
        deep: bool
            If True get a deep copy of the parameters.

        Returns
        -------
        dict[str, Any]
            Dictionary of parameter names and values.
        """
        parameters = super().get_params(deep)
        if deep:
            parameters["radius"] = copy.copy(self.radius)
            parameters["use_features"] = copy.copy(self.use_features)
        else:
            parameters["radius"] = self.radius
            parameters["use_features"] = self.use_features

        # remove fill_value from parameters
        parameters.pop("fill_value", None)
        return parameters

    def set_params(self, **parameters: Any) -> Self:
        """Set parameters.

        Parameters
        ----------
        parameters: Any
            Dictionary of parameter names and values.

        Returns
        -------
        Self
            PipelineElement with updated parameters.
        """
        parameter_copy = dict(parameters)
        radius = parameter_copy.pop("radius", None)
        use_features = parameter_copy.pop("use_features", None)

        # explicitly check for None, since 0 is a valid value
        if radius is not None:
            self._radius = radius
        # explicitly check for None, since False is a valid value
        if use_features is not None:
            self._use_features = bool(use_features)
        super().set_params(**parameter_copy)
        return self

    @property
    def radius(self) -> int:
        """Get radius of Morgan fingerprint."""
        return self._radius

    @property
    def use_features(self) -> bool:
        """Get whether to encode atoms by features or not."""
        return self._use_features

    @abc.abstractmethod
    def _explain_rdmol(self, mol_obj: RDKitMol) -> dict[int, list[tuple[int, int]]]:
        """Get central atom and radius of all features in molecule.

        Parameters
        ----------
        mol_obj: RDKitMol
            RDKit molecule to be encoded.

        Returns
        -------
        dict[int, list[tuple[int, int]]]
            Dictionary with mapping from bit to atom index and radius.
        """
        raise NotImplementedError

    def bit2atom_mapping(
        self, mol_obj: RDKitMol
    ) -> dict[int, list[CircularAtomEnvironment]]:
        """Obtain set of atoms for all features.

        Parameters
        ----------
        mol_obj: RDKitMol
            RDKit molecule to be encoded.

        Returns
        -------
        dict[int, list[CircularAtomEnvironment]]
            Dictionary with mapping from bit to encoded AtomEnvironments (which contain atom indices).
        """
        bit2atom_dict = self._explain_rdmol(mol_obj)
        result_dict: dict[int, list[CircularAtomEnvironment]] = {}
        # Iterating over all present bits and respective matches
        for bit, matches in bit2atom_dict.items():  # type: int, list[tuple[int, int]]
            result_dict[bit] = []
            for central_atom, radius in matches:  # type: int, int
                env = CircularAtomEnvironment.from_mol(mol_obj, central_atom, radius)
                result_dict[bit].append(env)
        # Transforming default dict to dict
        return result_dict



"""Unittest for handling Nones."""

import unittest
from typing import Any

import numpy as np
from rdkit import Chem, RDLogger
from rdkit.Chem.rdchem import MolSanitizeException
from sklearn.base import clone

from molpipeline import ErrorFilter, FilterReinserter, Pipeline, PostPredictionWrapper
from molpipeline.abstract_pipeline_elements.core import MolToMolPipelineElement
from molpipeline.any2mol import SmilesToMol
from molpipeline.any2mol.auto2mol import AutoToMol
from molpipeline.mol2any import MolToMorganFP, MolToRDKitPhysChem, MolToSmiles
from molpipeline.utils.molpipeline_types import OptionalMol, RDKitMol
from tests.utils.mock_element import MockTransformingPipelineElement

rdlog = RDLogger.logger()
rdlog.setLevel(RDLogger.CRITICAL)

TEST_SMILES = ["NCCCO", "abc", "c1ccccc1"]
EXPECTED_OUTPUT = ["NCCCO", None, "c1ccccc1"]


class NoneTest(unittest.TestCase):
    """Unittest for None Handling."""

    def test_error_dummy_fill_molpipeline(self) -> None:
        """Assert that invalid smiles are transformed to None."""

        smi2mol = SmilesToMol()
        mol2smi = MolToSmiles()
        remove_error = ErrorFilter.from_element_list([smi2mol, mol2smi])
        replace_error = PostPredictionWrapper(
            FilterReinserter.from_error_filter(remove_error, fill_value=None)
        )

        pipeline = Pipeline(
            [
                ("smi2mol", smi2mol),
                ("mol2smi", mol2smi),
                ("remove_error", remove_error),
                ("replace_error", replace_error),
            ]
        )
        out = pipeline.fit_transform(TEST_SMILES)
        for pred_val, true_val in zip(out, EXPECTED_OUTPUT):
            self.assertEqual(pred_val, true_val)

    def test_error_dummy_remove_record_molpipeline(self) -> None:
        """Assert that invalid smiles are transformed to None."""
        smi2mol = SmilesToMol()
        mol2smi = MolToSmiles()
        error_filter = ErrorFilter.from_element_list([smi2mol, mol2smi])
        pipeline = Pipeline(
            [
                ("smi2mol", smi2mol),
                ("mol2smi", mol2smi),
                ("error_filter", error_filter),
            ]
        )
        out = pipeline.transform(TEST_SMILES)
        self.assertEqual(len(out), 2)
        self.assertEqual(out[0], EXPECTED_OUTPUT[0])

    def test_dummy_remove_morgan_record_molpipeline(self) -> None:
        """Assert that invalid smiles are transformed to None."""
        smi2mol = SmilesToMol()
        mol2morgan = MolToMorganFP()
        error_filter = ErrorFilter.from_element_list([smi2mol, mol2morgan])
        pipeline = Pipeline(
            [
                ("smi2mol", smi2mol),
                ("mol2morgan", mol2morgan),
                ("error_filter", error_filter),
            ],
        )
        out = pipeline.transform(TEST_SMILES)
        self.assertEqual(out.shape, (2, 2048))

    def test_dummy_remove_physchem_record_molpipeline(self) -> None:
        """Assert that invalid smiles are transformed to None."""
        smi2mol = SmilesToMol()
        mol2physchem = MolToRDKitPhysChem()
        remove_none = ErrorFilter.from_element_list([smi2mol, mol2physchem])
        pipeline = Pipeline(
            [
                ("smi2mol", smi2mol),
                ("mol2physchem", mol2physchem),
                ("remove_none", remove_none),
            ],
        )
        pipeline2 = clone(pipeline)
        pipeline.fit(TEST_SMILES)
        out = pipeline.transform(TEST_SMILES)
        out2 = pipeline2.fit_transform(TEST_SMILES)
        self.assertEqual(out.shape, out2.shape)
        self.assertTrue(np.max(np.abs(out - out2)) < 0.000001)

    def test_dummy_remove_physchem_record_autodetect_molpipeline(self) -> None:
        """Assert that invalid smiles are transformed to None."""
        smi2mol = SmilesToMol()
        mol2physchem = MolToRDKitPhysChem()
        remove_none = ErrorFilter(filter_everything=True)
        pipeline = Pipeline(
            [
                ("smi2mol", smi2mol),
                ("mol2physchem", mol2physchem),
                ("remove_none", remove_none),
            ],
        )
        pipeline2 = clone(pipeline)
        pipeline.fit(TEST_SMILES)
        out = pipeline.transform(TEST_SMILES)
        print(pipeline2["remove_none"].filter_everything)
        out2 = pipeline2.fit_transform(TEST_SMILES)
        self.assertEqual(out.shape, out2.shape)
        self.assertTrue(np.max(np.abs(out - out2)) < 0.000001)

    def test_dummy_fill_physchem_record_molpipeline(self) -> None:
        """Assert that invalid smiles are transformed to None."""

        smi2mol = SmilesToMol()
        mol2physchem = MolToRDKitPhysChem()
        remove_none = ErrorFilter.from_element_list([smi2mol, mol2physchem])
        fill_none = PostPredictionWrapper(
            FilterReinserter.from_error_filter(remove_none, fill_value=10)
        )

        pipeline = Pipeline(
            [
                ("smi2mol", smi2mol),
                ("mol2physchem", mol2physchem),
                ("remove_none", remove_none),
                ("fill_none", fill_none),
            ],
            n_jobs=1,
        )
        pipeline2 = clone(pipeline)
        pipeline.fit(TEST_SMILES)
        out = pipeline.transform(TEST_SMILES)
        out2 = pipeline2.fit_transform(TEST_SMILES)
        self.assertEqual(out.shape, out2.shape)
        self.assertEqual(out.shape, (3, 215))
        self.assertTrue(np.nanmax(np.abs(out - out2)) < 0.000001)

    def test_replace_mixed_datatypes(self) -> None:
        """Assert that invalid values are replaced by fill value."""

        # test values
        test_values = [
            123,
            20099,  # index 1 will be replaced in this test
            1337,
        ]

        test_tuples = [
            (42, False),  # test scalar fill value with list returned
            (np.nan, False),  # test float fill value with list returned
            ([], False),  # test object list fill value with list returned
            ({}, False),  # test object dict fill value with list returned
            (None, False),  # test None fill value with list returned
            (42, True),  # test scalar fill value with numpy array returned
            (np.nan, True),  # test float fill value with numpy array returned
            ({}, True),  # test object dict fill value with numpy array returned
            (None, True),  # test None fill value with numpy array returned
        ]

        for fill_value, as_numpy_array in test_tuples:
            this_test_values: Any = test_values
            if as_numpy_array:
                this_test_values = np.array(test_values)

            mock2mock = MockTransformingPipelineElement(
                invalid_values={
                    test_values[1]
                },  # replaces element at index 1 with an invalid instance
                return_as_numpy_array=as_numpy_array,
            )
            error_filter = ErrorFilter.from_element_list([mock2mock])
            error_replacer = FilterReinserter.from_error_filter(
                error_filter=error_filter, fill_value=fill_value
            )
            pipeline = Pipeline(
                [
                    ("mock2mock", mock2mock),
                    ("error_filter", error_filter),
                    ("error_replacer", error_replacer),
                ],
                n_jobs=1,
            )
            pipeline2 = clone(pipeline)
            pipeline.fit(this_test_values)
            out = pipeline.transform(this_test_values)
            out2 = pipeline2.fit_transform(this_test_values)

            if as_numpy_array:
                self.assertTrue(isinstance(out, np.ndarray))
                self.assertTrue(isinstance(out2, np.ndarray))
            else:
                self.assertTrue(isinstance(out, list))
                self.assertTrue(isinstance(out2, list))

            self.assertEqual(len(out), len(out2))
            self.assertEqual(len(this_test_values), len(out))
            self.assertEqual(out[0], this_test_values[0])
            self.assertEqual(out2[0], this_test_values[0])
            self.assertEqual(out[2], this_test_values[2])
            self.assertEqual(out2[2], this_test_values[2])

            if isinstance(fill_value, float) and np.isnan(fill_value).all():
                self.assertTrue(np.isnan(out[1]).all())
                self.assertTrue(np.isnan(out2[1]).all())
            else:
                if fill_value is None:
                    self.assertIsNone(out[1])
                    self.assertIsNone(out2[1])
                else:
                    self.assertEqual(out[1], fill_value)
                    self.assertEqual(out2[1], fill_value)

    def test_replace_mixed_datatypes_expected_failures(self) -> None:
        """Test expected failures when replacing invalid values."""

        # test values
        test_values = [
            123,
            20099,  # index 1 will be replaced in this test
            1337,
        ]

        mock2mock = MockTransformingPipelineElement(
            invalid_values={test_values[1]},
            return_as_numpy_array=True,
        )
        error_filter = ErrorFilter.from_element_list([mock2mock])
        fill_value: list[Any] = []
        error_replacer = FilterReinserter.from_error_filter(
            error_filter=error_filter, fill_value=fill_value
        )
        pipeline = Pipeline(
            [
                ("mock2mock", mock2mock),
                ("error_filter", error_filter),
                ("error_replacer", error_replacer),
            ],
            n_jobs=1,
        )
        pipeline2 = clone(pipeline)

        self.assertRaises(ValueError, pipeline.fit, test_values)
        self.assertRaises(ValueError, pipeline.transform, test_values)
        self.assertRaises(ValueError, pipeline2.fit_transform, test_values)

    def test_molsanitize_error(self) -> None:
        """Test if MolSanitizeException is caught and catched by ErrorFilter."""

        class DummyMolSanitizeExc(MolToMolPipelineElement):
            """MolToMolPipelineElement with dummy molsanitize exception."""

            def pretransform_single(self, value: RDKitMol) -> OptionalMol:
                """Dummy Mol.

                Parameters
                ----------
                value: RDKitMol
                    Molecule.

                Returns
                -------
                OptionalMol
                    Molecule.
                """
                if Chem.MolToSmiles(value) == "c1ccccc1":
                    raise MolSanitizeException("This is a dummy exception.")
                return value

        pipeline = Pipeline(
            [
                ("autotosmiles", AutoToMol()),
                ("atomneutralizer", DummyMolSanitizeExc()),
                ("moltosmiles", MolToSmiles()),
                ("errorfilter", error_filter := ErrorFilter()),
                (
                    "filterreinserter",
                    FilterReinserter.from_error_filter(error_filter, None),
                ),
            ],
            n_jobs=-1,
        )

        result = pipeline.transform(["c1ccccc1", "CCCCCCC", "c1cc"])
        self.assertEqual(result, [None, "CCCCCCC", None])



"""Class for Transforming SDF-strings to rdkit molecules."""

from __future__ import annotations

from typing import Any, Optional

try:
    from typing import Self  # type: ignore[attr-defined]
except ImportError:
    from typing_extensions import Self

import copy

from rdkit import Chem

from molpipeline.abstract_pipeline_elements.any2mol.string2mol import (
    StringToMolPipelineElement as _StringToMolPipelineElement,
)
from molpipeline.abstract_pipeline_elements.core import InvalidInstance
from molpipeline.utils.molpipeline_types import OptionalMol


class SDFToMol(_StringToMolPipelineElement):
    """PipelineElement transforming a list of SDF strings to mol_objects."""

    identifier: str
    mol_counter: int

    # pylint: disable=R0913
    def __init__(
        self,
        identifier: str = "enumerate",
        name: str = "SDF2Mol",
        n_jobs: int = 1,
        uuid: Optional[str] = None,
    ) -> None:
        """Initialize SDFToMol.

        Parameters
        ----------
        identifier: str
            Method of assigning identifiers to molecules. At the moment molecules are counted.
        name: str
            Name of PipelineElement
        n_jobs: int
            Number of cores used for processing.
        uuid: Optional[str], optional
            uuid of PipelineElement, by default None
        """
        super().__init__(name=name, n_jobs=n_jobs, uuid=uuid)
        self.identifier = identifier
        self.mol_counter = 0

    def get_params(self, deep: bool = True) -> dict[str, Any]:
        """Return all parameters defining the object.

        Parameters
        ----------
        deep: bool
            If True get a deep copy of the parameters.

        Returns
        -------
        dict[str, Any]
            Dictionary containing all parameters defining the object.
        """
        params = super().get_params(deep)
        if deep:
            params["identifier"] = copy.copy(self.identifier)
        else:
            params["identifier"] = self.identifier
        return params

    def set_params(self, **parameters: Any) -> Self:
        """Set parameters of the object.

        Parameters
        ----------
        parameters: Any
            Dictionary containing all parameters defining the object.

        Returns
        -------
        Self
            SDFToMol with updated parameters.
        """
        super().set_params(**parameters)
        if "identifier" in parameters:
            self.identifier = parameters["identifier"]
        return self

    def finish(self) -> None:
        """Reset the mol counter which assigns identifiers."""
        self.mol_counter = 0

    def pretransform_single(self, value: str) -> OptionalMol:
        """Transform an SDF-strings to a rdkit molecule.

        Parameters
        ----------
        value: str
            SDF-string to transform to a molecule.

        Returns
        -------
        OptionalMol
            Molecule if transformation was successful, else InvalidInstance.
        """
        if not isinstance(value, (str, bytes)):
            return InvalidInstance(
                self.uuid,
                "Invalid SDF string!",
                self.name,
            )
        mol = Chem.MolFromMolBlock(value)
        if mol is None:
            return InvalidInstance(
                self.uuid,
                "Invalid SDF string!",
                self.name,
            )
        if self.identifier == "smiles":
            mol.SetProp("identifier", str(self.mol_counter))
        self.mol_counter += 1
        return mol



"""Classes for creating arrays from multiple concatenated descriptors or fingerprints."""

from __future__ import annotations

from typing import Any, Iterable, Optional, Union

try:
    from typing import Self  # type: ignore[attr-defined]
except ImportError:
    from typing_extensions import Self

import numpy as np
import numpy.typing as npt
from loguru import logger
from sklearn.base import clone

from molpipeline.abstract_pipeline_elements.core import (
    InvalidInstance,
    MolToAnyPipelineElement,
)
from molpipeline.abstract_pipeline_elements.mol2any.mol2bitvector import (
    MolToFingerprintPipelineElement,
)
from molpipeline.utils.molpipeline_types import RDKitMol


class MolToConcatenatedVector(MolToAnyPipelineElement):
    """Creates a concatenated descriptor vectored from multiple MolToAny PipelineElements."""

    _element_list: list[tuple[str, MolToAnyPipelineElement]]

    # pylint: disable=R0913
    def __init__(
        self,
        element_list: list[tuple[str, MolToAnyPipelineElement]],
        use_feature_names_prefix: bool = True,
        name: str = "MolToConcatenatedVector",
        n_jobs: int = 1,
        uuid: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        """Initialize MolToConcatenatedVector.

        Parameters
        ----------
        element_list: list[MolToAnyPipelineElement]
            List of Pipeline Elements of which the output is concatenated.
        use_feature_names_prefix: bool, optional (default=True)
            If True, will add the pipeline element's name as prefix to feature names.
            If False, only the feature names are used. This can lead to duplicate
            feature names.
        name: str, optional (default="MolToConcatenatedVector")
            name of pipeline.
        n_jobs: int, optional (default=1)
            Number of cores used.
        uuid: str | None, optional (default=None)
            UUID of the pipeline element. If None, a random UUID is generated.
        kwargs: Any
            Additional keyword arguments. Can be used to set parameters of the pipeline elements.
        """
        self._element_list = element_list
        if len(element_list) == 0:
            raise ValueError("element_list must contain at least one element.")
        self._use_feature_names_prefix = use_feature_names_prefix
        super().__init__(name=name, n_jobs=n_jobs, uuid=uuid)
        # set element execution details
        self._set_element_execution_details(self._element_list)
        # set feature names
        self._feature_names = self._create_feature_names(
            self._element_list, self._use_feature_names_prefix
        )
        self.set_params(**kwargs)

    @property
    def element_list(self) -> list[tuple[str, MolToAnyPipelineElement]]:
        """Get pipeline elements."""
        return self._element_list

    @property
    def n_features(self) -> int:
        """Calculates and returns the number of features."""
        feature_count = 0
        for _, element in self._element_list:
            if hasattr(element, "n_features"):
                feature_count += element.n_features
            elif hasattr(element, "n_bits"):
                feature_count += element.n_bits
            else:
                raise ValueError(
                    f"Element {element} does not have n_features or n_bits."
                )
        return feature_count

    @property
    def feature_names(self) -> list[str]:
        """Return the feature names of concatenated elements."""
        return self._feature_names[:]

    @staticmethod
    def _create_feature_names(
        element_list: list[tuple[str, MolToAnyPipelineElement]],
        use_feature_names_prefix: bool,
    ) -> list[str]:
        """Create feature names for concatenated vector from its elements.

        Parameters
        ----------
        element_list: list[tuple[str, MolToAnyPipelineElement]]
            List of pipeline elements.
        use_feature_names_prefix: bool
            If True, will add the pipeline element's name as prefix to feature names.
            If False, only the feature names are used. This can lead to duplicate
            feature names.

        Raises
        ------
        ValueError
            If element does not have feature_names attribute.

        Returns
        -------
        list[str]
            List of feature names.
        """
        feature_names = []
        for name, element in element_list:
            if not hasattr(element, "feature_names"):
                raise ValueError(
                    f"Element {element} does not have feature_names attribute."
                )

            if use_feature_names_prefix:
                # use element name as prefix
                feature_names.extend(
                    [f"{name}__{feature}" for feature in element.feature_names]  # type: ignore[attr-defined]
                )
            else:
                feature_names.extend(element.feature_names)  # type: ignore[attr-defined]

        if len(feature_names) != len(set(feature_names)):
            logger.warning(
                "Feature names in MolToConcatenatedVector are not unique."
                " Set use_feature_names_prefix=True and use unique pipeline element"
                " names to avoid this."
            )
        return feature_names

    def _set_element_execution_details(
        self, element_list: list[tuple[str, MolToAnyPipelineElement]]
    ) -> None:
        """Set output type and requires fitting for the concatenated vector.

        Parameters
        ----------
        element_list: list[tuple[str, MolToAnyPipelineElement]]
            List of pipeline elements.
        """
        output_types = set()
        for _, element in self._element_list:
            element.n_jobs = self.n_jobs
            output_types.add(element.output_type)
        if len(output_types) == 1:
            self._output_type = output_types.pop()
        else:
            self._output_type = "mixed"
        self._requires_fitting = any(
            element[1]._requires_fitting  # pylint: disable=protected-access
            for element in element_list
        )

    def get_params(self, deep: bool = True) -> dict[str, Any]:
        """Return all parameters defining the object.

        Parameters
        ----------
        deep: bool
            If True get a deep copy of the parameters.

        Returns
        -------
        dict[str, Any]
            Parameters defining the object.
        """
        parameters = super().get_params(deep)
        if deep:
            parameters["element_list"] = [
                (str(name), clone(ele)) for name, ele in self.element_list
            ]
            parameters["use_feature_names_prefix"] = bool(
                self._use_feature_names_prefix
            )
        else:
            parameters["element_list"] = self.element_list
            parameters["use_feature_names_prefix"] = self._use_feature_names_prefix
        for name, element in self.element_list:
            for key, value in element.get_params(deep=deep).items():
                parameters[f"{name}__{key}"] = value

        return parameters

    def _set_element_list(
        self, parameter_copy: dict[str, Any], **parameters: Any
    ) -> tuple[dict[str, Any], dict[str, Any]]:
        """Set the element list and run necessary configurations.

        Parameters
        ----------
        parameter_copy: dict[str, Any]
            Copy of parameters.
        parameters: Any
            Original parameters.

        Raises
        ------
        ValueError
            If element_list is empty.

        Returns
        -------
        tuple[dict[str, Any], dict[str, Any]]
            Updated parameter_copy and parameters.
        """
        element_list = parameter_copy.pop("element_list", None)
        if element_list is not None:
            self._element_list = element_list
            if len(element_list) == 0:
                raise ValueError("element_list must contain at least one element.")
            # reset element execution details
            self._set_element_execution_details(self._element_list)
        step_params: dict[str, dict[str, Any]] = {}
        step_dict = dict(self._element_list)
        to_delete_list = []
        for parm, value in parameters.items():
            if "__" not in parm:
                continue
            param_split = parm.split("__")
            param_header = param_split[0]
            # Check if parameter addresses an element
            if param_header not in step_dict:
                continue
            param_tail = "__".join(param_split[1:])
            if param_header not in step_params:
                step_params[param_header] = {}
            step_params[param_header][param_tail] = value
            to_delete_list.append(parm)
        for to_delete in to_delete_list:
            _ = parameter_copy.pop(to_delete, None)
        for step, params in step_params.items():
            step_dict[step].set_params(**params)
        return parameter_copy, parameters

    def set_params(self, **parameters: Any) -> Self:
        """Set parameters.

        Parameters
        ----------
        parameters: Any
            Parameters to set.

        Returns
        -------
        Self
            Mol2ConcatenatedVector object with updated parameters.
        """
        parameter_copy = dict(parameters)

        # handle element_list
        parameter_copy, parameters = self._set_element_list(
            parameter_copy, **parameters
        )

        # handle use_feature_names_prefix
        use_feature_names_prefix = parameter_copy.pop("use_feature_names_prefix", None)
        if use_feature_names_prefix is not None:
            self._use_feature_names_prefix = use_feature_names_prefix
            # reset feature names
            self._feature_names = self._create_feature_names(
                self._element_list,
                self._use_feature_names_prefix,  # type: ignore[arg-type]
            )

        # set parameters of super
        super().set_params(**parameter_copy)
        return self

    def assemble_output(
        self,
        value_list: Iterable[npt.NDArray[np.float64]],
    ) -> npt.NDArray[np.float64]:
        """Transform output of all transform_single operations to matrix.

        Parameters
        ----------
        value_list: Iterable[npt.NDArray[np.float64]]
            List of molecular descriptors or fingerprints which are concatenated to a single matrix.

        Returns
        -------
        npt.NDArray[np.float64]
            Matrix of shape (n_molecules, n_features) with concatenated features specified during init.
        """
        return np.vstack(list(value_list))

    def transform(self, values: list[RDKitMol]) -> npt.NDArray[np.float64]:
        """Transform the list of molecules to sparse matrix.

        Parameters
        ----------
        values: list[RDKitMol]
            List of molecules to transform.

        Returns
        -------
        npt.NDArray[np.float64]
            Matrix of shape (n_molecules, n_features) with concatenated features specified during init.
        """
        output: npt.NDArray[np.float64] = super().transform(values)
        return output

    def fit(
        self,
        values: list[RDKitMol],
        labels: Any = None,  # pylint: disable=unused-argument
    ) -> Self:
        """Fit each pipeline element.

        Parameters
        ----------
        values: list[RDKitMol]
            List of molecules used to fit the pipeline elements creating the concatenated vector.
        labels: Any
            Labels for the molecules. Not used.

        Returns
        -------
        Self
            Fitted pipeline element.
        """
        for pipeline_element in self._element_list:
            pipeline_element[1].fit(values)
        return self

    def pretransform_single(
        self, value: RDKitMol
    ) -> Union[list[Union[npt.NDArray[np.float64], dict[int, int]]], InvalidInstance]:
        """Get pretransform of each element and concatenate for output.

        Parameters
        ----------
        value: RDKitMol
            Molecule to be transformed.

        Returns
        -------
        Union[list[Union[npt.NDArray[np.float64], dict[int, int]]], InvalidInstance]
            List of pretransformed values of each pipeline element.
            If any element returns None, InvalidInstance is returned.
        """
        final_vector = []
        error_message = ""
        for name, pipeline_element in self._element_list:
            vector = pipeline_element.pretransform_single(value)
            if isinstance(vector, InvalidInstance):
                error_message += f"{self.name}__{name} returned an InvalidInstance."
                break

            final_vector.append(vector)
        else:  # no break
            return final_vector
        return InvalidInstance(self.uuid, error_message, self.name)

    def finalize_single(self, value: Any) -> Any:
        """Finalize the output of transform_single.

        Parameters
        ----------
        value: Any
            Output of transform_single.

        Returns
        -------
        Any
            Finalized output.
        """
        final_vector_list = []
        for (_, element), sub_value in zip(self._element_list, value):
            final_value = element.finalize_single(sub_value)
            if isinstance(element, MolToFingerprintPipelineElement) and isinstance(
                final_value, dict
            ):
                vector = np.zeros(element.n_bits)
                vector[list(final_value.keys())] = np.array(list(final_value.values()))
                final_value = vector
            if not isinstance(final_value, np.ndarray):
                final_value = np.array(final_value)
            final_vector_list.append(final_value)
        return np.hstack(final_vector_list)

    def fit_to_result(self, values: Any) -> Self:
        """Fit the pipeline element to the result of transform_single.

        Parameters
        ----------
        values: Any
            Output of transform_single.

        Returns
        -------
        Self
            Fitted pipeline element.
        """
        for element, value in zip(self._element_list, zip(*values)):
            element[1].fit_to_result(value)
        return self



"""Test visualization methods for explanations."""

import unittest
from typing import ClassVar

import numpy as np
from rdkit import Chem
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor

from molpipeline import Pipeline
from molpipeline.any2mol import SmilesToMol
from molpipeline.experimental.explainability import (
    SHAPFeatureAndAtomExplanation,
    SHAPFeatureExplanation,
    SHAPKernelExplainer,
    SHAPTreeExplainer,
    structure_heatmap,
    structure_heatmap_shap,
)
from molpipeline.mol2any import MolToMorganFP
from tests.test_experimental.test_explainability.utils import (
    construct_kernel_shap_kwargs,
)

TEST_SMILES = ["CC", "CCO", "COC", "c1ccccc1(N)", "CCC(-O)O", "CCCN"]
CONTAINS_OX = [0, 1, 1, 0, 1, 0]  # classification labels
REGRESSION_LABELS = [0.1, 0.2, 0.3, 0.4, 0.5, 0.6]  # regression labels


_RANDOM_STATE = 67056


def _get_test_morgan_rf_pipeline(task: str = "classification") -> Pipeline:
    """Get a test pipeline with Morgan fingerprints and a random forest classifier.

    Parameters
    ----------
    task : str, optional (default="classification")
        Task of the pipeline. Either "classification" or "regression".

    Returns
    -------
    Pipeline
        Pipeline with Morgan fingerprints and a random forest classifier.
    """

    if task == "classification":
        model = RandomForestClassifier(n_estimators=2, random_state=_RANDOM_STATE)
    elif task == "regression":
        model = RandomForestRegressor(n_estimators=2, random_state=_RANDOM_STATE)
    else:
        raise ValueError(f"Invalid task: {task}")

    pipeline = Pipeline(
        [
            ("smi2mol", SmilesToMol()),
            ("morgan", MolToMorganFP(radius=1, n_bits=1024)),
            (
                "model",
                model,
            ),
        ]
    )
    return pipeline


class TestExplainabilityVisualization(unittest.TestCase):
    """Test the public interface of the visualization methods for explanations."""

    test_pipeline_clf: ClassVar[Pipeline]
    test_tree_explainer_clf: ClassVar[SHAPTreeExplainer]
    test_tree_explanations_clf: ClassVar[
        list[SHAPFeatureAndAtomExplanation | SHAPFeatureExplanation]
    ]
    test_kernel_explainer_clf: ClassVar[SHAPKernelExplainer]
    test_kernel_explanations_clf: ClassVar[
        list[SHAPFeatureAndAtomExplanation | SHAPFeatureExplanation]
    ]

    test_pipeline_reg: ClassVar[Pipeline]
    test_tree_explainer_reg: ClassVar[SHAPTreeExplainer]
    test_tree_explanations_reg: ClassVar[
        list[SHAPFeatureAndAtomExplanation | SHAPFeatureExplanation]
    ]
    test_kernel_explainer_reg: ClassVar[SHAPKernelExplainer]
    test_kernel_explanations_reg: ClassVar[
        list[SHAPFeatureAndAtomExplanation | SHAPFeatureExplanation]
    ]

    @classmethod
    def setUpClass(cls) -> None:
        """Set up the tests."""
        # test pipeline for classification
        cls.test_pipeline_clf = _get_test_morgan_rf_pipeline(task="classification")
        cls.test_pipeline_clf.fit(TEST_SMILES, CONTAINS_OX)

        # test pipeline for regression
        cls.test_pipeline_reg = _get_test_morgan_rf_pipeline(task="regression")
        cls.test_pipeline_reg.fit(TEST_SMILES, REGRESSION_LABELS)

        # tree explainer for classification
        cls.test_tree_explainer_clf = SHAPTreeExplainer(cls.test_pipeline_clf)
        cls.test_tree_explanations_clf = cls.test_tree_explainer_clf.explain(
            TEST_SMILES,
        )

        # tree explainer for regression
        cls.test_tree_explainer_reg = SHAPTreeExplainer(cls.test_pipeline_reg)
        cls.test_tree_explanations_reg = cls.test_tree_explainer_reg.explain(
            TEST_SMILES
        )

        # kernel explainer for classification
        kernel_kwargs_clf = construct_kernel_shap_kwargs(
            cls.test_pipeline_clf, TEST_SMILES
        )
        cls.test_kernel_explainer_clf = SHAPKernelExplainer(
            cls.test_pipeline_clf, **kernel_kwargs_clf
        )
        cls.test_kernel_explanations_clf = cls.test_kernel_explainer_clf.explain(
            TEST_SMILES
        )

        # kernel explainer for regression
        kernel_kwargs_reg = construct_kernel_shap_kwargs(
            cls.test_pipeline_reg, TEST_SMILES
        )
        cls.test_kernel_explainer_reg = SHAPKernelExplainer(
            cls.test_pipeline_reg, **kernel_kwargs_reg
        )
        cls.test_kernel_explanations_reg = cls.test_kernel_explainer_reg.explain(
            TEST_SMILES
        )

    def test_structure_heatmap_fingerprint_based_atom_coloring(self) -> None:
        """Test structure heatmap fingerprint-based atom coloring."""
        for explanation_list in [
            self.test_tree_explanations_clf,
            self.test_kernel_explanations_clf,
            self.test_tree_explanations_reg,
            self.test_kernel_explanations_reg,
        ]:
            for explanation in explanation_list:
                self.assertTrue(explanation.is_valid())
                self.assertIsInstance(explanation.atom_weights, np.ndarray)  # type: ignore[union-attr]
                image = structure_heatmap(
                    explanation.molecule,
                    explanation.atom_weights,  # type: ignore
                    width=8,
                    height=8,
                )  # type: ignore[union-attr]
                self.assertIsNotNone(image)
                self.assertEqual(image.format, "PNG")

    def test_structure_heatmap_shap_explanation(self) -> None:
        """Test structure heatmap SHAP explanation."""
        for explanation_list in [
            self.test_tree_explanations_clf,
            self.test_kernel_explanations_clf,
            self.test_tree_explanations_reg,
            self.test_kernel_explanations_reg,
        ]:
            for explanation in explanation_list:
                self.assertTrue(explanation.is_valid())
                self.assertIsInstance(explanation, SHAPFeatureAndAtomExplanation)
                self.assertIsInstance(explanation.atom_weights, np.ndarray)  # type: ignore[union-attr]
                image = structure_heatmap_shap(
                    explanation=explanation,  # type: ignore[arg-type]
                    width=8,
                    height=8,
                )  # type: ignore[union-attr]
                self.assertIsNotNone(image)
                self.assertEqual(image.format, "PNG")

    def test_explicit_hydrogens(self) -> None:
        """Test that the visualization methods work with explicit hydrogens."""
        mol_implicit_hydrogens = Chem.MolFromSmiles("C")
        explanations1 = self.test_tree_explainer_clf.explain(
            [Chem.MolToSmiles(mol_implicit_hydrogens)]
        )
        mol_added_hydrogens = Chem.AddHs(mol_implicit_hydrogens)
        explanations2 = self.test_tree_explainer_clf.explain(
            [Chem.MolToSmiles(mol_added_hydrogens)]
        )
        mol_explicit_hydrogens = Chem.MolFromSmiles("[H]C([H])([H])[H]")
        explanations3 = self.test_tree_explainer_clf.explain(
            [Chem.MolToSmiles(mol_explicit_hydrogens)]
        )

        # test explanations' atom weights
        self.assertEqual(len(explanations1), 1)
        self.assertEqual(len(explanations2), 1)
        self.assertEqual(len(explanations3), 1)
        self.assertTrue(hasattr(explanations1[0], "atom_weights"))
        self.assertTrue(hasattr(explanations2[0], "atom_weights"))
        self.assertTrue(hasattr(explanations3[0], "atom_weights"))
        self.assertIsInstance(explanations1[0].atom_weights, np.ndarray)  # type: ignore[union-attr]
        self.assertIsInstance(explanations2[0].atom_weights, np.ndarray)  # type: ignore[union-attr]
        self.assertIsInstance(explanations3[0].atom_weights, np.ndarray)  # type: ignore[union-attr]
        self.assertEqual(len(explanations1[0].atom_weights), 1)  # type: ignore
        self.assertEqual(len(explanations2[0].atom_weights), 1)  # type: ignore
        self.assertEqual(len(explanations3[0].atom_weights), 1)  # type: ignore

        # test visualization
        all_explanations = explanations1 + explanations2 + explanations3
        for explanation in all_explanations:
            self.assertTrue(explanation.is_valid())
            image = structure_heatmap(
                explanation.molecule,
                explanation.atom_weights,  # type: ignore
                width=8,
                height=8,
            )  # type: ignore[union-attr]
            self.assertIsNotNone(image)
            self.assertEqual(image.format, "PNG")



"""Initialize the molpipeline package."""

# pylint: disable=no-name-in-module
from rdkit.Chem import PropertyPickleOptions, SetDefaultPickleProperties

from molpipeline.error_handling import ErrorFilter, FilterReinserter
from molpipeline.pipeline import Pipeline
from molpipeline.post_prediction import PostPredictionWrapper

# Keep all properties when pickling. Otherwise, we will lose properties set on RDKitMol when passed to
# multiprocessing subprocesses.
SetDefaultPickleProperties(PropertyPickleOptions.AllProps)

__all__ = [
    "__version__",
    "Pipeline",
    "ErrorFilter",
    "FilterReinserter",
    "PostPredictionWrapper",
]

__version__ = "0.10.1"



"""Classes ment to transform given input to a RDKit molecule."""

from __future__ import annotations

from rdkit import Chem

from molpipeline.abstract_pipeline_elements.core import (
    AnyToMolPipelineElement,
    InvalidInstance,
)
from molpipeline.utils.molpipeline_types import OptionalMol


class BinaryToMol(AnyToMolPipelineElement):
    """Transforms binary string representation to RDKit Mol objects."""

    def pretransform_single(self, value: str) -> OptionalMol:
        """Transform binary string to molecule.

        Parameters
        ----------
        value: str
            Binary string.

        Returns
        -------
        OptionalMol
            Rdkit molecule if valid binary representation, else None.
        """
        if value is None:
            return InvalidInstance(
                self.uuid,
                f"Invalid binary string: {value}",
                self.name,
            )

        if not isinstance(value, bytes):
            return InvalidInstance(
                self.uuid,
                f"Not bytes: {value}",
                self.name,
            )

        mol: OptionalMol | None = None
        try:
            mol = Chem.Mol(value)
        except RuntimeError:
            pass

        if not mol:
            return InvalidInstance(
                self.uuid,
                f"Invalid binary string: {value}",
                self.name,
            )
        return mol



"""Module for elements that are applied in the pipeline after the prediction."""

from __future__ import annotations

import abc
from typing import Any, Optional, Union

try:
    from typing import Self  # type: ignore[attr-defined]
except ImportError:
    from typing_extensions import Self

from numpy import typing as npt
from sklearn.base import BaseEstimator, TransformerMixin

from molpipeline.abstract_pipeline_elements.core import ABCPipelineElement
from molpipeline.error_handling import FilterReinserter
from molpipeline.utils.molpipeline_types import AnyPredictor, AnyTransformer


class PostPredictionTransformation(BaseEstimator, TransformerMixin, abc.ABC):
    """Abstract base class for post prediction steps.

    Post prediction steps are used to transform the output after the prediction step.
    E.g. dummy values are added for entries removed during preprocessing.
    """

    @abc.abstractmethod
    def transform(self, X: Any, **params: Any) -> Any:  # pylint: disable=invalid-name
        """Transform data.

        Parameters
        ----------
        X: npt.NDArray[Any]
            Input data.
        **params: Any
            Additional parameters for transformation.

        Returns
        -------
        npt.NDArray[Any]
            Transformed data.
        """


class PostPredictionWrapper(PostPredictionTransformation):
    """Wrapper for post prediction transformations.

    This class is used to wrap a PipelineElement in a PostPredictionTransformation.
    """

    def __init__(
        self,
        wrapped_estimator: Union[AnyPredictor, AnyTransformer, FilterReinserter[Any]],
        **kwargs: Any,
    ) -> None:
        """Initialize PostPredictionWrapper.

        Parameters
        ----------
        wrapped_estimator: PipelineElement
            PipelineElement to be wrapped.
        kwargs: Any
            Parameter of the wrapped_estimator
        Returns
        -------
        None
        """
        self.wrapped_estimator = wrapped_estimator
        if kwargs:
            if isinstance(self.wrapped_estimator, ABCPipelineElement):
                self.wrapped_estimator.set_params(**kwargs)
            else:
                self.wrapped_estimator.set_params(**kwargs)

    def fit(
        self,
        X: npt.NDArray[Any],  # pylint: disable=invalid-name
        y: Optional[npt.NDArray[Any]] = None,  # pylint: disable=invalid-name
        **params: Any,
    ) -> Self:
        """Fit PostPredictionWrapper.

        Parameters
        ----------
        X : npt.NDArray[Any]
            Input data.
        y : Optional[npt.NDArray[Any]]
            Target data.
        **params : Any
            Additional parameters for fitting.

        Returns
        -------
        Self
            Fitted PostPredictionWrapper.
        """
        if isinstance(self.wrapped_estimator, FilterReinserter):
            self.wrapped_estimator.fit(X, **params)
        else:
            self.wrapped_estimator.fit(X, y, **params)
        return self

    def transform(
        self,
        X: npt.NDArray[Any],  # pylint: disable=invalid-name
        **params: Any,
    ) -> npt.NDArray[Any]:
        """Transform data.

        Parameters
        ----------
        X : npt.NDArray[Any]
            Input data.
        **params : Any
            Additional parameters for transformation.

        Returns
        -------
        npt.NDArray[Any]
            Transformed data.
        """
        if hasattr(self.wrapped_estimator, "predict"):
            return self.wrapped_estimator.predict(X, **params)
        if hasattr(self.wrapped_estimator, "transform"):
            return self.wrapped_estimator.transform(X, **params)
        raise AttributeError(
            f"Estimator {self.wrapped_estimator} has neither predict nor transform method."
        )

    def fit_transform(
        self,
        X: npt.NDArray[Any],
        y: Optional[npt.NDArray[Any]] = None,
        **params: Any,
    ) -> npt.NDArray[Any]:
        """Fit and transform data.

        Parameters
        ----------
        X: npt.NDArray[Any]
            Input data.
        y: npt.NDArray[Any]
            Target data.
        **params: Any
            Additional parameters for fitting.

        Returns
        -------
        npt.NDArray[Any]
            Transformed data.
        """
        if hasattr(self.wrapped_estimator, "fit_predict"):
            return self.wrapped_estimator.fit_predict(X, y, **params)
        if hasattr(self.wrapped_estimator, "fit_transform"):
            if isinstance(self.wrapped_estimator, FilterReinserter):
                return self.wrapped_estimator.fit_transform(X)
            return self.wrapped_estimator.fit_transform(X, y, **params)
        raise AttributeError(
            f"Estimator {self.wrapped_estimator} has neither fit_predict nor fit_transform method."
        )

    def inverse_transform(
        self, X: npt.NDArray[Any]  # pylint: disable=invalid-name
    ) -> npt.NDArray[Any]:
        """Inverse transform data.

        Parameters
        ----------
        X: npt.NDArray[Any]
            Input data.

        Returns
        -------
        npt.NDArray[Any]
            Inverse transformed data.
        """
        if hasattr(self.wrapped_estimator, "inverse_transform"):
            return self.wrapped_estimator.inverse_transform(X)
        raise AttributeError(
            f"Estimator {self.wrapped_estimator} has no inverse_transform method."
        )

    def get_params(self, deep: bool = True) -> dict[str, Any]:
        """Get parameters.

        Parameters
        ----------
        deep: bool
            Whether to recursively get parameters.

        Returns
        -------
        dict[str, Any]
            Parameters.
        """
        param_dict = {"wrapped_estimator": self.wrapped_estimator}
        if deep:
            for key, value in self.wrapped_estimator.get_params(deep=deep).items():
                param_dict[f"wrapped_estimator__{key}"] = value
        return param_dict

    def set_params(self, **params: Any) -> Self:
        """Set parameters.

        Parameters
        ----------
        **params: Any
            Parameters to set.

        Returns
        -------
        dict[str, Any]
            Parameters.
        """
        param_copy = dict(params)
        if "wrapped_estimator" in param_copy:
            self.wrapped_estimator = param_copy.pop("wrapped_estimator")
        wrapped_estimator_params = {}
        for key, value in param_copy.items():
            estimator, _, param = key.partition("__")
            if estimator == "wrapped_estimator":
                wrapped_estimator_params[param] = value
        self.wrapped_estimator.set_params(**wrapped_estimator_params)
        return self



"""Test Chemprop component wrapper."""

import logging
import unittest
from typing import Iterable

import torch
from chemprop.nn.loss import MSELoss
from sklearn.base import clone

from molpipeline.estimators.chemprop.component_wrapper import (
    MPNN,
    BondMessagePassing,
    MeanAggregation,
    MulticlassClassificationFFN,
    RegressionFFN,
    SumAggregation,
)
from molpipeline.estimators.chemprop.models import (
    ChempropClassifier,
    ChempropModel,
    ChempropMulticlassClassifier,
    ChempropRegressor,
)
from molpipeline.estimators.chemprop.neural_fingerprint import ChempropNeuralFP
from molpipeline.utils.json_operations import recursive_from_json, recursive_to_json

# pylint: disable=relative-beyond-top-level
from test_extras.test_chemprop.chemprop_test_utils.compare_models import compare_params
from test_extras.test_chemprop.chemprop_test_utils.constant_vars import (
    DEFAULT_BINARY_CLASSIFICATION_PARAMS,
    DEFAULT_MULTICLASS_CLASSIFICATION_PARAMS,
    DEFAULT_SET_PARAMS,
    NO_IDENTITY_CHECK,
)
from test_extras.test_chemprop.chemprop_test_utils.default_models import (
    get_chemprop_model_binary_classification_mpnn,
)

logging.getLogger("lightning.pytorch.utilities.rank_zero").setLevel(logging.WARNING)


class TestChempropModel(unittest.TestCase):
    """Test the Chemprop model."""

    def test_get_params(self) -> None:
        """Test the get_params and set_params methods."""
        chemprop_model = get_chemprop_model_binary_classification_mpnn()
        orig_params = chemprop_model.get_params(deep=True)
        expected_params = dict(DEFAULT_BINARY_CLASSIFICATION_PARAMS)  # Shallow copy

        self.assertSetEqual(set(orig_params), set(expected_params))
        # Check if the parameters are as expected
        for param_name, param in expected_params.items():
            if param_name in NO_IDENTITY_CHECK:
                if isinstance(param, Iterable):
                    self.assertIsInstance(orig_params[param_name], type(param))
                    for i, p in enumerate(param):
                        self.assertIsInstance(orig_params[param_name][i], p)
                elif isinstance(param, type):
                    self.assertIsInstance(orig_params[param_name], param)
                else:
                    raise ValueError(f"{param_name} should be a type.")
            else:
                self.assertEqual(
                    orig_params[param_name], param, f"Test failed for {param_name}"
                )

        new_params = {
            "batch_size": 32,
            "model__agg": MeanAggregation(),
            "model__message_passing__activation": "tanh",
            "model__message_passing__depth": 2,
        }
        # Check setting new parameters
        chemprop_model.set_params(**new_params)
        model_params = chemprop_model.get_params(deep=True)
        for param_name, param in new_params.items():
            if param_name in ["model__agg"]:
                self.assertIsInstance(model_params[param_name], type(param))
                continue
            self.assertEqual(param, model_params[param_name])

    def test_clone(self) -> None:
        """Test the clone method."""
        chemprop_model = get_chemprop_model_binary_classification_mpnn()
        cloned_model = clone(chemprop_model)
        self.assertIsInstance(cloned_model, ChempropModel)
        compare_params(self, chemprop_model, cloned_model)

    def test_classifier_methods(self) -> None:
        """Test the classifier methods."""
        chemprop_model = get_chemprop_model_binary_classification_mpnn()
        # pylint: disable=protected-access
        self.assertTrue(chemprop_model._is_binary_classifier())
        self.assertFalse(chemprop_model._is_multiclass_classifier())
        # pylint: enable=protected-access
        self.assertTrue(hasattr(chemprop_model, "predict_proba"))

    def test_neural_fp(self) -> None:
        """Test the to_encoder method."""
        chemprop_model = get_chemprop_model_binary_classification_mpnn()
        neural_fp = chemprop_model.to_encoder()
        self.assertIsInstance(neural_fp, ChempropNeuralFP)
        self.assertIsInstance(neural_fp.model, MPNN)
        # the model should be cloned
        self.assertNotEqual(id(chemprop_model.model), id(neural_fp.model))
        self.assertEqual(neural_fp.disable_fitting, True)

    def test_json_serialization(self) -> None:
        """Test the to_json and from_json methods."""
        chemprop_model = get_chemprop_model_binary_classification_mpnn()
        chemprop_json = recursive_to_json(chemprop_model)
        chemprop_model_copy = recursive_from_json(chemprop_json)
        param_dict = chemprop_model_copy.get_params(deep=True)

        self.assertSetEqual(
            set(param_dict.keys()), set(DEFAULT_BINARY_CLASSIFICATION_PARAMS.keys())
        )
        for param_name, param in DEFAULT_BINARY_CLASSIFICATION_PARAMS.items():
            if param_name in NO_IDENTITY_CHECK:
                if isinstance(param, Iterable):
                    self.assertIsInstance(param_dict[param_name], type(param))
                    for i, p in enumerate(param):
                        self.assertIsInstance(param_dict[param_name][i], p)
                elif isinstance(param, type):
                    self.assertIsInstance(param_dict[param_name], param)
                else:
                    raise ValueError(f"{param_name} should be a type.")
            elif param_name == "model__predictor__task_weights":
                self.assertTrue(torch.allclose(param, param_dict[param_name]))
            else:
                self.assertEqual(
                    param_dict[param_name], param, f"Test failed for {param_name}"
                )


class TestChempropClassifier(unittest.TestCase):
    """Test the Chemprop classifier model."""

    def test_get_params(self) -> None:
        """Test the get_params and set_params methods."""
        chemprop_model = ChempropClassifier(lightning_trainer__accelerator="cpu")
        param_dict = chemprop_model.get_params(deep=True)
        expected_params = dict(DEFAULT_BINARY_CLASSIFICATION_PARAMS)  # Shallow copy
        self.assertSetEqual(set(param_dict.keys()), set(expected_params.keys()))
        for param_name, param in expected_params.items():
            if param_name in NO_IDENTITY_CHECK:
                if isinstance(param, Iterable):
                    self.assertIsInstance(param_dict[param_name], type(param))
                    for i, p in enumerate(param):
                        self.assertIsInstance(param_dict[param_name][i], p)
                elif isinstance(param, type):
                    self.assertIsInstance(param_dict[param_name], param)
                else:
                    raise ValueError(f"{param_name} should be a type.")
            else:
                self.assertEqual(
                    param_dict[param_name], param, f"Test failed for {param_name}"
                )

    def test_set_params(self) -> None:
        """Test the set_params methods."""
        chemprop_model = ChempropClassifier(lightning_trainer__accelerator="cpu")
        chemprop_model.set_params(**DEFAULT_SET_PARAMS)
        current_params = chemprop_model.get_params(deep=True)
        for param, value in DEFAULT_SET_PARAMS.items():
            self.assertEqual(current_params[param], value)


class TestChempropRegressor(unittest.TestCase):
    """Test the Chemprop regressor model."""

    def test_get_params(self) -> None:
        """Test the get_params and set_params methods."""
        chemprop_model = ChempropRegressor(lightning_trainer__accelerator="cpu")
        param_dict = chemprop_model.get_params(deep=True)
        expected_params = dict(DEFAULT_BINARY_CLASSIFICATION_PARAMS)
        expected_params["model__predictor"] = RegressionFFN
        expected_params["model__predictor__criterion"] = MSELoss
        self.assertSetEqual(set(param_dict.keys()), set(expected_params.keys()))
        for param_name, param in expected_params.items():
            if param_name in NO_IDENTITY_CHECK:
                if isinstance(param, Iterable):
                    self.assertIsInstance(param_dict[param_name], type(param))
                    for i, p in enumerate(param):
                        self.assertIsInstance(param_dict[param_name][i], p)
                elif isinstance(param, type):
                    self.assertIsInstance(param_dict[param_name], param)
                else:
                    raise ValueError(f"{param_name} should be a type.")
            else:
                self.assertEqual(
                    param_dict[param_name], param, f"Test failed for {param_name}"
                )


class TestChempropMulticlassClassifier(unittest.TestCase):
    """Test the Chemprop classifier model."""

    def test_get_params(self) -> None:
        """Test the get_params and set_params methods."""
        n_classes = 3
        chemprop_model = ChempropMulticlassClassifier(
            lightning_trainer__accelerator="cpu", n_classes=n_classes
        )
        param_dict = chemprop_model.get_params(deep=True)
        expected_params = dict(DEFAULT_MULTICLASS_CLASSIFICATION_PARAMS)  # Shallow copy
        expected_params["model__predictor__n_classes"] = n_classes
        expected_params["n_classes"] = n_classes
        self.assertSetEqual(set(param_dict.keys()), set(expected_params.keys()))
        for param_name, param in expected_params.items():
            if param_name in NO_IDENTITY_CHECK:
                if isinstance(param, Iterable):
                    self.assertIsInstance(param_dict[param_name], type(param))
                    for i, p in enumerate(param):
                        self.assertIsInstance(param_dict[param_name][i], p)
                elif isinstance(param, type):
                    self.assertIsInstance(param_dict[param_name], param)
                else:
                    raise ValueError(f"{param_name} should be a type.")
            elif isinstance(param, torch.Tensor):
                self.assertTrue(torch.allclose(param_dict[param_name], param))
            else:
                self.assertEqual(
                    param_dict[param_name], param, f"Test failed for {param_name}"
                )

    def test_set_params(self) -> None:
        """Test the set_params methods."""
        chemprop_model = ChempropMulticlassClassifier(
            lightning_trainer__accelerator="cpu", n_classes=3
        )
        chemprop_model.set_params(**DEFAULT_SET_PARAMS)
        params = {
            "n_classes": 4,
            "batch_size": 20,
            "lightning_trainer__max_epochs": 10,
            "model__predictor__n_layers": 2,
        }
        chemprop_model.set_params(**params)
        current_params = chemprop_model.get_params(deep=True)
        for param, value in params.items():
            self.assertEqual(current_params[param], value)

    def test_error_for_multiclass_predictor(self) -> None:
        """Test the error for using a multiclass predictor for a binary classification model."""
        bond_encoder = BondMessagePassing()
        agg = SumAggregation()
        with self.assertRaises(ValueError):
            predictor = MulticlassClassificationFFN(n_classes=2)
            model = MPNN(message_passing=bond_encoder, agg=agg, predictor=predictor)
            ChempropMulticlassClassifier(n_classes=2, model=model)
        with self.assertRaises(ValueError):
            predictor = MulticlassClassificationFFN(n_classes=3)
            model = MPNN(message_passing=bond_encoder, agg=agg, predictor=predictor)
            ChempropMulticlassClassifier(n_classes=4, model=model)
        with self.assertRaises(AttributeError):
            predictor = RegressionFFN()
            model = MPNN(message_passing=bond_encoder, agg=agg, predictor=predictor)
            ChempropMulticlassClassifier(n_classes=4, model=model)



"""Contains functions for loading and saving pipeline elements and models to json files."""

from __future__ import annotations

import types
import typing
import warnings
from typing import Any

from molpipeline.pipeline import Pipeline
from molpipeline.utils.json_operations_torch import tensor_to_json

__all__ = [
    "builtin_to_json",
    "decode_dict",
    "recursive_from_json",
    "recursive_to_json",
    "transform_functions2string",
    "transform_string2function",
]


def transform_functions2string(value: Any) -> Any:
    """Transform functions to string representation.

    If the value is a function, it is transformed to a dictionary containing the module and the class name.
    If the value is a dictionary, the function is called recursively for each value.
    If the value is a list, the function is called recursively for each value.
    Else the value is returned as is.

    Parameters
    ----------
    value: Any
        Value which is transformed.

    Returns
    -------
    Any
        Json file containing the dictionary.
    """
    if callable(value):
        out_dict = {
            "load_from_constructor": True,
            "__name__": value.__name__,
            "__module__": value.__module__,
        }
        return out_dict

    if isinstance(value, dict):
        out_dict = {}
        for dict_key, dict_value in value.items():
            out_dict[dict_key] = transform_functions2string(dict_value)
        return out_dict

    if isinstance(value, list):
        out_list = []
        for list_value in value:
            out_list.append(transform_functions2string(list_value))
        return out_list

    return value


def transform_string2function(value: Any) -> Any:
    """Transform string representation of functions to actual functions.

    If the value is a dictionary containing the key "load_from_constructor" and the value is True,
    the function is loaded from the module and class name.
    If the value is a dictionary, the function is called recursively for each value.
    If the value is a list, the function is called recursively for each value.
    Else the value is returned as is.

    Parameters
    ----------
    value: Any
        Object to be transformed

    Returns
    -------
    Any
        Json file containing the dictionary.
    """
    if isinstance(value, dict):
        if "load_from_constructor" in value:
            if value["load_from_constructor"]:
                module_str: str = value["__module__"]
                class_str: str = value["__name__"]
                class_module = __import__(module_str, fromlist=[class_str])
                return getattr(class_module, class_str)
            return value
        out_dict = {}
        for dict_key, dict_value in value.items():
            out_dict[dict_key] = transform_string2function(dict_value)
        return out_dict

    if isinstance(value, list):
        out_list = []
        for list_value in value:
            out_list.append(transform_string2function(list_value))
        return out_list

    return value


_U = typing.TypeVar("_U", str, int, float, bool, None)


@typing.overload
def builtin_to_json(obj: _U) -> _U:
    """Transform a builtin object to an object savable as json file.

    Parameters
    ----------
    obj: str | int | float | bool | None
        Object which would be transformed, but these types are just returned as is.

    Returns
    -------
    str | int | float | bool | None
        The same object as the input.
    """


@typing.overload
def builtin_to_json(obj: list[Any]) -> list[Any]:
    """Transform a builtin object to an object savable as json file.

    Parameters
    ----------
    obj: list[Any]
        List of objects transformed recursively to json compatible objects.

    Returns
    -------
    list[Any]
        List of transformed objects.
    """


@typing.overload
def builtin_to_json(obj: tuple[Any, ...]) -> tuple[Any, ...]:
    """Transform a builtin object to an object savable as json file.

    Parameters
    ----------
    obj: tuple[Any]
        Tuple of objects transformed recursively to json compatible objects.

    Returns
    -------
    tuple[Any]
        Tuple of transformed objects.
    """


@typing.overload
def builtin_to_json(
    obj: types.FunctionType | set[Any] | dict[Any, Any],
) -> dict[str, Any]:
    """Transform a builtin object to an object savable as json file.

    Parameters
    ----------
    obj: types.FunctionType | set[Any] | dict[Any, Any]
        Object which are encoded as a dictionary in order to be json compatible.

    Returns
    -------
    tuple[Any]
        Tuple of transformed objects.
    """


def builtin_to_json(obj: Any) -> Any:
    """Transform a builtin object to an object savable as json file.

    Parameters
    ----------
    obj: PythonNative
        Object to be transformed. Can be a string, int, float, bool, list, tuple, dict, callable or a set.

    Returns
    -------
    Any
        Json file containing the dictionary.
    """
    if isinstance(obj, (str, int, float, bool)) or obj is None:
        return obj

    if isinstance(obj, types.FunctionType):
        return {
            "__name__": obj.__name__,
            "__module__": obj.__module__,
            "__init__": False,
        }
    if isinstance(obj, dict):
        return {
            recursive_to_json(key): recursive_to_json(value)
            for key, value in obj.items()
        }
    if isinstance(obj, (list, tuple)):
        iter_list = [recursive_to_json(value) for value in obj]
        iterable_type = type(obj)
        return iterable_type(iter_list)

    object_dict: dict[str, Any] = {
        "__name__": obj.__class__.__name__,
        "__module__": obj.__class__.__module__,
        "__init__": True,
    }
    # If the object is a sklearn model, the parameters for initialization are extracted.
    if isinstance(obj, set):
        object_dict["__set_items__"] = [recursive_to_json(value) for value in obj]
        return object_dict
    raise TypeError(f"Unexpected Type: {type(obj)}")


def recursive_to_json(obj: Any) -> Any:
    """Recursively transform an object to a json file.

    Parameters
    ----------
    obj: Any
        Object to be transformed. Can be a string, int, float, bool, list, tuple, dict, callable or a sklearn model.
        A sklearn model is defined as an object with a get_params method.

    Returns
    -------
    dict[str, Any]
        Json file containing the dictionary.
    """
    if obj is None:
        return None

    if isinstance(
        obj, (str, int, float, bool, dict, types.FunctionType, list, set, tuple)
    ):
        return_value = builtin_to_json(obj)
        return return_value

    # If neither of the above, it is assumed to be an object.
    object_dict: dict[str, Any] = {
        "__name__": obj.__class__.__name__,
        "__module__": obj.__class__.__module__,
        "__init__": True,
    }
    # If the object is a sklearn model, the parameters for initialization are extracted.
    if hasattr(obj, "get_params"):
        if isinstance(obj, Pipeline):
            object_dict["steps"] = [
                (step_name, recursive_to_json(step_model))
                for (step_name, step_model) in obj.steps
            ]
            object_dict["memory"] = obj.memory
            object_dict["verbose"] = obj.verbose
            object_dict["n_jobs"] = obj.n_jobs
        else:
            model_params = obj.get_params(deep=False)
            for key, value in model_params.items():
                object_dict[key] = recursive_to_json(value)
    else:
        obj_dict, success = tensor_to_json(obj)
        # Either not a tensor or torch is not available
        if success:
            return obj_dict
        # If the object is not a sklearn model, a warning is raised
        # as it might not be possible to recreate the object.
        warnings.warn(
            f"{type(obj)} has no get_params method. No parameters for initialization are retained."
        )

    return object_dict


def decode_dict(obj: dict[str, Any]) -> Any:
    """Decode a dictionary to an object.

    Parameters
    ----------
    obj: dict[str, Any]
        Dictionary to be transformed

    Returns
    -------
    Any
        Object specified in the dictionary.
    """
    # Create copy
    object_params_copy = dict(obj)

    # For functions of classes
    obj_module_str = object_params_copy.pop("__module__", None)
    obj_class_str = object_params_copy.pop("__name__", None)
    initialize = object_params_copy.pop("__init__", False)

    # Convert remaining Values
    converted_dict = {}
    for key, value in object_params_copy.items():
        converted_dict[key] = recursive_from_json(value)

    # If the object is a function or a class
    if obj_module_str and obj_class_str:
        class_module = __import__(obj_module_str, fromlist=[obj_class_str])
        obj_class = getattr(class_module, obj_class_str)
        if not initialize:  # If the object is a function or should not be initialized
            return obj_class
        # If the object is a class, but has no parameters
        if not converted_dict:
            return obj_class()
        if obj_class is set:
            return set(converted_dict["__set_items__"])
        # If the object is a class, and has parameters
        return obj_class(**converted_dict)
    return converted_dict


def recursive_from_json(obj: Any) -> Any:
    """Recursively transform a json file to an object.

    Parameters
    ----------
    obj: Any
        Object to be transformed

    Returns
    -------
    Any
        Object specified in the json file.
    """
    if isinstance(obj, (str, int, float, bool)) or obj is None:
        return obj

    if isinstance(obj, dict):
        return decode_dict(obj)

    if isinstance(obj, (list, tuple)):
        iter_list = [recursive_from_json(value) for value in obj]
        iterable_type = type(obj)
        return iterable_type(iter_list)

    raise TypeError(f"Unexpected Type: {type(obj)}")



"""All abstract classes later pipeline elements inherit from."""

from __future__ import annotations  # for all the python 3.8 users out there.

import abc
import copy
import inspect
from typing import Any, Iterable, NamedTuple, Optional, Union

try:
    from typing import Self  # type: ignore[attr-defined]
except ImportError:
    from typing_extensions import Self

from uuid import uuid4

import numpy as np
from joblib import Parallel, delayed
from loguru import logger
from rdkit import Chem
from rdkit.Chem import Mol as RDKitMol  # pylint: disable=no-name-in-module

from molpipeline.utils.multi_proc import check_available_cores


class InvalidInstance(NamedTuple):
    """Object which is returned when an instance cannot be processed.

    Attributes
    ----------
    element_id: str
        Id of the element which could not be processed.
    message: str
        Message why the element could not be processed.
    element_name: Optional[str]
        Optional name of the element which could not be processed.
        The name of the pipeline element is often more descriptive than the id.
    """

    element_id: str
    message: str
    element_name: Optional[str] = None

    def __repr__(self) -> str:
        """Return string representation of InvalidInstance.

        Returns
        -------
        str
            String representation of InvalidInstance.
        """
        return (
            f"InvalidInstance({self.element_name or self.element_id}, {self.message})"
        )


OptionalMol = Union[RDKitMol, InvalidInstance]


class RemovedInstance:  # pylint: disable=too-few-public-methods
    """Object which is returned by a ErrorFilter if an Invalid instance was removed."""

    def __init__(self, filter_element_id: str, message: Optional[str] = None) -> None:
        """Initialize RemovedInstance.

        Parameters
        ----------
        filter_element_id: str
            FilterElement which removed the molecule.
        message: Optional[str]
            Optional message why the molecule was removed.

        Returns
        -------
        None
        """
        self.filter_element_id = filter_element_id
        self.message = message

    def __repr__(self) -> str:
        """Return string representation of RemovedInstance.

        Returns
        -------
        str
            String representation of RemovedInstance.
        """
        return f"RemovedInstance({self.filter_element_id}, {self.message})"


class ABCPipelineElement(abc.ABC):
    """Ancestor of all PipelineElements."""

    name: str
    _requires_fitting: bool = False
    uuid: str

    def __init__(
        self,
        name: Optional[str] = None,
        n_jobs: int = 1,
        uuid: Optional[str] = None,
    ) -> None:
        """Initialize ABCPipelineElement.

        Parameters
        ----------
        name: Optional[str], optional (default=None)
            Name of PipelineElement
        n_jobs: int
            Number of cores used for processing.
        uuid: Optional[str]
            Unique identifier of the PipelineElement.
        """
        if name is None:
            name = self.__class__.__name__
        self.name = name
        self.n_jobs = n_jobs
        if uuid is None:
            self.uuid = str(uuid4())
        else:
            self.uuid = uuid

    def __repr__(self) -> str:
        """Return string representation of object.

        Returns
        -------
        str
            String representation of object.
        """
        parm_list = []
        for key, value in self._get_non_default_parameters().items():
            parm_list.append(f"{key}={value}")
        parm_str = ", ".join(parm_list)
        repr_str = f"{self.__class__.__name__}({parm_str})"
        return repr_str

    def _get_non_default_parameters(self) -> dict[str, Any]:
        """Return all parameters which are not default.

        Returns
        -------
        dict[str, Any]
            Dictionary of non default parameters.
        """
        signature = inspect.signature(self.__class__.__init__)
        self_params = self.get_params()
        non_default_params = {}
        for parm, value_default in signature.parameters.items():
            if parm == "uuid":
                continue
            if value_default is inspect.Parameter.empty:
                non_default_params[parm] = self_params[parm]
                continue
            if parm in self_params:
                if value_default.default != self_params[parm]:
                    non_default_params[parm] = self_params[parm]
        return non_default_params

    def get_params(self, deep: bool = True) -> dict[str, Any]:
        """Return the parameters of the object.

        Parameters
        ----------
        deep: bool
            If True get a deep copy of the parameters.

        Returns
        -------
        dict[str, Any]
            Parameters of the object.
        """
        if deep:
            return {
                "name": copy.copy(self.name),
                "n_jobs": copy.copy(self.n_jobs),
                "uuid": copy.copy(self.uuid),
            }

        return {
            "name": self.name,
            "n_jobs": self.n_jobs,
            "uuid": self.uuid,
        }

    def set_params(self, **parameters: Any) -> Self:
        """As the setter function cannot be assessed with super(), this method is implemented for inheritance.

        Parameters
        ----------
        parameters: Any
            Parameters to be set.

        Returns
        -------
        Self
            Self with updated parameters.
        """
        for att_name, att_value in parameters.items():
            if not hasattr(self, att_name):
                raise ValueError(
                    f"Cannot set attribute {att_name} on {self.__class__.__name__}"
                )
            setattr(self, att_name, att_value)
        return self

    @property
    def additional_attributes(self) -> dict[str, Any]:
        """Any attribute relevant for recreating and exact copy, which is not a parameter."""
        return {}

    @property
    def n_jobs(self) -> int:
        """Get the number of cores."""
        return self._n_jobs

    @n_jobs.setter
    def n_jobs(self, n_jobs: int) -> None:
        """Set the number of cores.

        Parameters
        ----------
        n_jobs: int
            Number of cores used for processing.

        Returns
        -------
        None
        """
        self._n_jobs = check_available_cores(n_jobs)

    @property
    def requires_fitting(self) -> bool:
        """Return whether the object requires fitting or not."""
        return self._requires_fitting

    def finish(self) -> None:
        """Inform object that iteration has been finished. Does in most cases nothing.

        Called after all transform singles have been processed. From MolPipeline
        """

    def fit(self, values: Any, labels: Any = None) -> Self:
        """Fit object to input_values.

        Most objects might not need fitting, but it is implemented for consitency for all PipelineElements.

        Parameters
        ----------
        values: Any
            List of molecule representations.
        labels: Any
            Optional label for fitting.

        Returns
        -------
        Self
            Fitted object.
        """
        _ = self.fit_transform(values, labels)
        return self

    def fit_to_result(self, values: Any) -> Self:  # pylint: disable=unused-argument
        """Fit object to result of transformed values.

        Fit object to the result of the transform function. This is useful catching nones and removed molecules.

        Parameters
        ----------
        values: Any
            List of molecule representations.

        Returns
        -------
        Self
            Fitted object.
        """
        return self

    @abc.abstractmethod
    def fit_transform(
        self,
        values: Any,
        labels: Any = None,
    ) -> Any:
        """Apply fit function and subsequently transform the input.

        Parameters
        ----------
        values: Any
            Apply transformation specified in transform_single to all molecules in the value_list.
        labels: Any
            Optional label for fitting.

        Returns
        -------
        Any
            List of instances in new representation.
        """

    @abc.abstractmethod
    def transform(self, values: Any) -> Any:
        """Transform input_values according to object rules.

        Parameters
        ----------
        values: Any
            Iterable of molecule representations (SMILES, MolBlocks RDKit Molecules, PhysChem vectors etc.).
            Input depends on the concrete PipelineElement.

        Returns
        -------
        Any
            Transformed input_values.
        """

    @abc.abstractmethod
    def transform_single(self, value: Any) -> Any:
        """Transform a single value.

        Parameters
        ----------
        value: Any
            Value to be transformed.

        Returns
        -------
        Any
            Transformed value.
        """


class TransformingPipelineElement(ABCPipelineElement):
    """Ancestor of all PipelineElements."""

    _input_type: str
    _output_type: str
    name: str

    def __init__(
        self,
        name: Optional[str] = None,
        n_jobs: int = 1,
        uuid: Optional[str] = None,
    ) -> None:
        """Initialize ABCPipelineElement.

        Parameters
        ----------
        name: Optional[str], optional (default=None)
            Name of PipelineElement
        n_jobs: int
            Number of cores used for processing.
        uuid: Optional[str]
            Unique identifier of the PipelineElement.
        """
        super().__init__(name=name, n_jobs=n_jobs, uuid=uuid)
        self._is_fitted = False

    @property
    def input_type(self) -> str:
        """Return the input type."""
        return self._input_type

    @property
    def is_fitted(self) -> bool:
        """Return whether the object is fitted or not."""
        return self._is_fitted

    @property
    def output_type(self) -> str:
        """Return the output type."""
        return self._output_type

    @property
    def parameters(self) -> dict[str, Any]:
        """Return the parameters of the object."""
        return self.get_params()

    @parameters.setter
    def parameters(self, **parameters: Any) -> None:
        """Set the parameters of the object.

        Parameters
        ----------
        parameters: Any
            Object parameters as a dictionary.

        Returns
        -------
        None
        """
        self.set_params(**parameters)

    def copy(self) -> Self:
        """Copy the object.

        Returns
        -------
        Self
            Copy of the object.
        """
        recreated_object = self.__class__(**self.parameters)
        for key, value in self.additional_attributes.items():
            if not hasattr(recreated_object, key):
                raise AssertionError(
                    f"Cannot set attribute {key} on {self.__class__.__name__}. This should not happen!"
                )
            setattr(recreated_object, key, copy.copy(value))
        return recreated_object

    def fit_to_result(self, values: Any) -> Self:
        """Fit object to result of transformed values.

        Fit object to the result of the transform function. This is useful catching nones and removed molecules.

        Parameters
        ----------
        values: Any
            List of molecule representations.

        Returns
        -------
        Self
            Fitted object.
        """
        self._is_fitted = True
        return super().fit_to_result(values)

    def fit_transform(
        self,
        values: Any,
        labels: Any = None,
    ) -> Any:
        """Apply fit function and subsequently transform the input.

        Parameters
        ----------
        values: Any
            Apply transformation specified in transform_single to all molecules in the value_list.
        labels: Any
            Optional label for fitting.

        Returns
        -------
        Any
            List of molecules in new representation.
        """
        self._is_fitted = True
        if self.requires_fitting:
            pre_value_list = self.pretransform(values)
            self.fit_to_result(pre_value_list)
            output_list = self.finalize_list(pre_value_list)
            if hasattr(self, "assemble_output"):
                return self.assemble_output(output_list)
        return self.transform(values)

    def transform_single(self, value: Any) -> Any:
        """Transform a single molecule to the new representation.

        RemovedMolecule objects are passed without change, as no transformations are applicable.

        Parameters
        ----------
        value: Any
            Current representation of the molecule. (Eg. SMILES, RDKit Mol, ...)

        Returns
        -------
        Any
            New representation of the molecule. (Eg. SMILES, RDKit Mol, Descriptor-Vector, ...)
        """
        if isinstance(value, InvalidInstance):
            return value
        pre_value = self.pretransform_single(value)
        if isinstance(pre_value, InvalidInstance):
            return pre_value
        return self.finalize_single(pre_value)

    def assemble_output(self, value_list: Iterable[Any]) -> Any:
        """Aggregate rows, which in most cases is just return the list.

        Some representations might be better representd as a single object. For example a list of vectors can
        be transformed to a matrix.

        Parameters
        ----------
        value_list: Iterable[Any]
            Iterable of transformed rows.

        Returns
        -------
        Any
            Aggregated output. This can also be the original input.
        """
        return list(value_list)

    @abc.abstractmethod
    def pretransform_single(self, value: Any) -> Any:
        """Transform the instance, but skips parameters learned during fitting.

        This is the first step for the full transformation.
        It is followed by the finalize_single method and assemble output which collects all single transformations.
        These functions are split as they need to be accessed separately from outside the object.

        Parameters
        ----------
        value: Any
            Value to be pretransformed.

        Returns
        -------
        Any
            Pretransformed value. (Skips applying parameters learned during fitting)
        """

    def finalize_single(self, value: Any) -> Any:
        """Apply parameters learned during fitting to a single instance.

        Parameters
        ----------
        value: Any
            Value obtained from pretransform_single.

        Returns
        -------
        Any
            Finalized value.
        """
        return value

    def pretransform(self, value_list: Iterable[Any]) -> list[Any]:
        """Transform input_values according to object rules without fitting specifics.

        Parameters
        ----------
        value_list: Iterable[Any]
            Iterable of instances to be pretransformed.

        Returns
        -------
        list[Any]
            Transformed input_values.
        """
        parallel = Parallel(n_jobs=self.n_jobs)
        output_values = parallel(
            delayed(self.pretransform_single)(value) for value in value_list
        )
        return output_values

    def finalize_list(self, value_list: Iterable[Any]) -> list[Any]:
        """Transform list of values according to parameters learned during fitting.

        Parameters
        ----------
        value_list:  Iterable[Any]
            List of values to be transformed.

        Returns
        -------
        list[Any]
            List of transformed values.
        """
        parallel = Parallel(n_jobs=self.n_jobs)
        output_values = parallel(
            delayed(self.finalize_single)(value) for value in value_list
        )
        return output_values

    def transform(self, values: Any) -> Any:
        """Transform input_values according to object rules.

        Parameters
        ----------
        values: Any
            Iterable of molecule representations (SMILES, MolBlocks RDKit Molecules, PhysChem vectors etc.).
            Input depends on the concrete PipelineElement.

        Returns
        -------
        Any
            Transformed input_values.
        """
        output_rows = self.pretransform(values)
        output_rows = self.finalize_list(output_rows)
        output = self.assemble_output(output_rows)
        self.finish()
        return output

    def to_json(self) -> dict[str, Any]:
        """Return all defining attributes of object as dict.

        Returns
        -------
        dict[str, Any]
            A dictionary with all attributes necessary to initialize a object with same parameters.
        """
        json_dict: dict[str, Any] = {
            "__name__": self.__class__.__name__,
            "__module__": self.__class__.__module__,
        }
        json_dict.update(self.parameters)
        if self.additional_attributes:
            adittional_attributes = {}
            for key, value in self.additional_attributes.items():
                if isinstance(value, np.ndarray):
                    adittional_attributes[key] = value.tolist()
                else:
                    adittional_attributes[key] = value
            json_dict["additional_attributes"] = adittional_attributes
        return json_dict


class MolToMolPipelineElement(TransformingPipelineElement, abc.ABC):
    """Abstract PipelineElement where input and outputs are molecules."""

    _input_type = "RDKitMol"
    _output_type = "RDKitMol"

    def transform(self, values: list[OptionalMol]) -> list[OptionalMol]:
        """Transform list of molecules to list of molecules.

        Parameters
        ----------
        values: list[OptionalMol]
            List of molecules to be transformed.

        Returns
        -------
        list[OptionalMol]
            List of molecules or InvalidInstances, if corresponding transformation was not successful.
        """
        mol_list: list[OptionalMol] = super().transform(values)  # Stupid mypy...
        return mol_list

    def transform_single(self, value: OptionalMol) -> OptionalMol:
        """Apply pretransform_single and finalize_single in one step.

        Parameters
        ----------
        value: OptionalMol
            Molecule to be transformed.

        Returns
        -------
        OptionalMol
            Transformed molecule if transformation was successful, else InvalidInstance.
        """
        try:
            return super().transform_single(value)
        except Exception as exc:
            if isinstance(value, Chem.Mol):
                logger.error(
                    f"Failed to process: {Chem.MolToSmiles(value)} | ({self.name}) ({self.uuid})"
                )
            else:
                logger.error(
                    f"Failed to process: {value} ({type(value)}) | ({self.name}) ({self.uuid})"
                )
            raise exc

    @abc.abstractmethod
    def pretransform_single(self, value: RDKitMol) -> OptionalMol:
        """Transform the molecule to another molecule object.

        Do not apply parameters learned during fitting.

        Parameters
        ----------
        value: RDKitMol
            Molecule to be transformed.

        Returns
        -------
        OptionalMol
            Transformed molecule if transformation was successful, else InvalidInstance.
        """


class AnyToMolPipelineElement(TransformingPipelineElement, abc.ABC):
    """Abstract PipelineElement which creates molecules from different inputs."""

    _output_type = "RDKitMol"

    def transform(self, values: Any) -> list[OptionalMol]:
        """Transform list of instances to list of molecules.

        Parameters
        ----------
        values: Any
            Instances to be transformed to a list of molecules.

        Returns
        -------
        list[OptionalMol]
            List of molecules or InvalidInstances, if corresponding representation was invalid.
        """
        mol_list: list[OptionalMol] = super().transform(values)  # Stupid mypy...
        return mol_list

    @abc.abstractmethod
    def pretransform_single(self, value: Any) -> OptionalMol:
        """Transform the instance to a molecule, but skip parameters learned during fitting.

        Parameters
        ----------
        value: Any
            Representation to be transformed to a molecule.

        Returns
        -------
        OptionalMol
            Obtained molecule if valid representation, else InvalidInstance.
        """


class MolToAnyPipelineElement(TransformingPipelineElement, abc.ABC):
    """Abstract PipelineElement which creates molecules from different inputs."""

    _input_type = "RDKitMol"

    @abc.abstractmethod
    def pretransform_single(self, value: RDKitMol) -> Any:
        """Transform the molecule, but skip parameters learned during fitting.

        Parameters
        ----------
        value: RDKitMOl
            Molecule to be transformed.

        Returns
        -------
        Any
            Transformed molecule.
        """



"""Classes for encoding molecules as phys-chem vector."""

# pylint: disable=too-many-arguments

from __future__ import annotations

from typing import Any, Callable, Optional, Union

try:
    from typing import Self  # type: ignore[attr-defined]
except ImportError:
    from typing_extensions import Self

import copy

import numpy as np
import numpy.typing as npt
from loguru import logger
from rdkit import Chem, rdBase
from rdkit.Chem import Descriptors
from sklearn.preprocessing import StandardScaler

from molpipeline.abstract_pipeline_elements.core import InvalidInstance
from molpipeline.abstract_pipeline_elements.mol2any.mol2floatvector import (
    MolToDescriptorPipelineElement,
)
from molpipeline.utils.molpipeline_types import AnyTransformer, RDKitMol

RDKIT_DESCRIPTOR_DICT: dict[str, Callable[[Chem.Mol], float]]
RDKIT_DESCRIPTOR_DICT = dict(Descriptors.descList)

# MolWt is removed as ExactMolWt is already included.
# Ipc is removed because it causes trouble with numpy.
DEFAULT_DESCRIPTORS = [
    name for name in RDKIT_DESCRIPTOR_DICT if name not in ["MolWt", "Ipc"]
]


class MolToRDKitPhysChem(MolToDescriptorPipelineElement):
    """PipelineElement for creating a Descriptor vector based on RDKit phys-chem properties."""

    _descriptor_list: list[str]

    def __init__(
        self,
        descriptor_list: Optional[list[str]] = None,
        return_with_errors: bool = False,
        standardizer: Optional[AnyTransformer] = StandardScaler(),
        log_exceptions: bool = True,
        name: str = "Mol2RDKitPhysChem",
        n_jobs: int = 1,
        uuid: Optional[str] = None,
    ) -> None:
        """Initialize MolToRDKitPhysChem.

        Parameters
        ----------
        descriptor_list: Optional[list[str]], optional (default=None)
            List of descriptor names to calculate. If None, DEFAULT_DESCRIPTORS are used.
        return_with_errors: bool, optional (default = False)
            If False, the pipeline element will return an InvalidInstance if any error occurs during calculations.
            If True, the pipeline element will return a vector with NaN values for failed descriptor calculations.
        standardizer: Optional[AnyTransformer], optional (default=StandardScaler())
            Standardizer to use.
        log_exceptions: bool, optional (default=True)
            If True, traceback of exceptions occurring during descriptor calculation are logged.
        name: str, optional (default="Mol2RDKitPhysChem")
            Name of the PipelineElement.
        n_jobs: int, optional (default=1)
            Number of jobs to use for parallelization.
        uuid: Optional[str], optional (default=None)
            UUID of the PipelineElement. If None, a new UUID is generated.
        """
        self.descriptor_list = descriptor_list  # type: ignore
        self._feature_names = self._descriptor_list
        self._return_with_errors = return_with_errors
        self._log_exceptions = log_exceptions
        super().__init__(
            standardizer=standardizer,
            name=name,
            n_jobs=n_jobs,
            uuid=uuid,
        )

    @property
    def n_features(self) -> int:
        """Return the number of features."""
        return len(self._descriptor_list)

    @property
    def descriptor_list(self) -> list[str]:
        """Return a copy of the descriptor list. Alias of `feature_names`."""
        return self._descriptor_list[:]

    @descriptor_list.setter
    def descriptor_list(self, descriptor_list: list[str] | None) -> None:
        """Set the descriptor list.

        Parameters
        ----------
        descriptor_list: list[str] | None
            List of descriptor names to calculate. If None, DEFAULT_DESCRIPTORS are used.

        Raises
        ------
        ValueError
            If an unknown descriptor name is used.
        """
        if descriptor_list is None or descriptor_list is DEFAULT_DESCRIPTORS:
            # if None or DEFAULT_DESCRIPTORS are used, set the default descriptors
            self._descriptor_list = DEFAULT_DESCRIPTORS
        else:
            # check all user defined descriptors are valid
            for descriptor_name in descriptor_list:
                if descriptor_name not in RDKIT_DESCRIPTOR_DICT:
                    raise ValueError(
                        f"Unknown descriptor function with name: {descriptor_name}"
                    )
            self._descriptor_list = descriptor_list

    def pretransform_single(
        self, value: RDKitMol
    ) -> Union[npt.NDArray[np.float64], InvalidInstance]:
        """Transform a single molecule to a descriptor vector.

        Parameters
        ----------
        value: RDKitMol
            RDKit molecule to transform.

        Returns
        -------
        Optional[npt.NDArray[np.float64]]
            Descriptor vector for given molecule. None if calculation failed.
        """
        vec = np.full((len(self._descriptor_list),), np.nan)
        log_block = rdBase.BlockLogs()  # pylint: disable=unused-variable
        for i, name in enumerate(self._descriptor_list):
            descriptor_func = RDKIT_DESCRIPTOR_DICT[name]
            try:
                vec[i] = descriptor_func(value)
            except Exception:  # pylint: disable=broad-except
                if self._log_exceptions:
                    logger.exception(f"Failed calculating descriptor: {name}")
        del log_block
        if not self._return_with_errors and np.any(np.isnan(vec)):
            return InvalidInstance(self.uuid, "NaN in descriptor vector", self.name)
        return vec

    def get_params(self, deep: bool = True) -> dict[str, Any]:
        """Get the parameters of the pipeline element.

        Parameters
        ----------
        deep: bool
            If true create a deep copy of the parameters

        Returns
        -------
        dict[str, Any]
            Parameter of the pipeline element.
        """
        parent_dict = dict(super().get_params(deep=deep))
        if deep:
            parent_dict["descriptor_list"] = copy.deepcopy(self._descriptor_list)
            parent_dict["return_with_errors"] = copy.deepcopy(self._return_with_errors)
            parent_dict["log_exceptions"] = copy.deepcopy(self._log_exceptions)
        else:
            parent_dict["descriptor_list"] = self._descriptor_list
            parent_dict["return_with_errors"] = self._return_with_errors
            parent_dict["log_exceptions"] = self._log_exceptions
        return parent_dict

    def set_params(self, **parameters: Any) -> Self:
        """Set parameters.

        Parameters
        ----------
        parameters: Any
            Parameters to set

        Returns
        -------
        Self
            Self
        """
        parameters_shallow_copy = dict(parameters)
        params_list = ["descriptor_list", "return_with_errors", "log_exceptions"]
        for param_name in params_list:
            if param_name in parameters:
                setattr(self, f"_{param_name}", parameters[param_name])
                parameters_shallow_copy.pop(param_name)
        super().set_params(**parameters_shallow_copy)
        return self



"""Init."""

from molpipeline.any2mol.auto2mol import AutoToMol
from molpipeline.any2mol.bin2mol import BinaryToMol
from molpipeline.any2mol.inchi2mol import InchiToMol
from molpipeline.any2mol.sdf2mol import SDFToMol
from molpipeline.any2mol.smiles2mol import SmilesToMol

__all__ = [
    "AutoToMol",
    "BinaryToMol",
    "SmilesToMol",
    "InchiToMol",
    "SDFToMol",
]



"""Tests conversion of sklearn models to json and back."""

import unittest

from sklearn.ensemble import RandomForestClassifier
from sklearn.svm import SVC

from molpipeline import Pipeline
from molpipeline.utils.json_operations import (
    recursive_from_json,
    recursive_to_json,
    transform_functions2string,
    transform_string2function,
)
from molpipeline.utils.multi_proc import check_available_cores


class JsonConversionTest(unittest.TestCase):
    """Unittest for conversion of sklearn models to json and back."""

    def test_rf_reconstruction(self) -> None:
        """Test if the sklearn-rf can be reconstructed from json.

        Returns
        -------
        None
        """
        random_forest = RandomForestClassifier(n_estimators=200)
        recreated_rf = recursive_from_json(recursive_to_json(random_forest))
        self.assertEqual(random_forest.get_params(), recreated_rf.get_params())

    def test_svc_reconstruction(self) -> None:
        """Test if the sklearn-svc can be reconstructed from json.

        Returns
        -------
        None
        """
        svc = SVC()
        recreated_svc = recursive_from_json(recursive_to_json(svc))
        self.assertEqual(svc.get_params(), recreated_svc.get_params())

    def test_pipeline_reconstruction(self) -> None:
        """Test if the sklearn-pipleine can be reconstructed from json.

        Returns
        -------
        None
        """
        random_forest = RandomForestClassifier(n_estimators=200)
        svc = SVC()
        pipeline = Pipeline([("rf", random_forest), ("svc", svc)])
        recreated_pipeline = recursive_from_json(recursive_to_json(pipeline))

        original_params = pipeline.get_params()
        recreated_params = recreated_pipeline.get_params()
        original_steps = original_params.pop("steps")
        recreated_steps = recreated_params.pop("steps")

        # Separate comparison of the steps as models cannot be compared directly
        for (orig_name, orig_obj), (recreated_name, recreated_obj) in zip(
            original_steps, recreated_steps
        ):
            # Remove the model from the original params
            del original_params[orig_name]
            del recreated_params[recreated_name]
            self.assertEqual(orig_name, recreated_name)
            self.assertEqual(orig_obj.get_params(), recreated_obj.get_params())
            self.assertEqual(type(orig_obj), type(recreated_obj))
        self.assertEqual(original_params, recreated_params)

    def test_function_dict_json(self) -> None:
        """Test if a dict with objects can be reconstructed from json.

        Returns
        -------
        None
        """
        function_dict = {
            "dummy1": {"check_available_cores": check_available_cores},
            "dummy2": str,
            "dummy3": 1,
            "dummy4": [check_available_cores, check_available_cores, "test"],
        }
        function_json = transform_functions2string(function_dict)
        recreated_function_dict = transform_string2function(function_json)
        self.assertEqual(function_dict, recreated_function_dict)

    def test_set_transformation(self) -> None:
        """Test if a set can be reconstructed from json.

        Returns
        -------

        """
        test_set = {1, "a", (1, "a")}
        test_set_json = recursive_to_json(test_set)
        recreated_set = recursive_from_json(test_set_json)
        self.assertEqual(test_set, recreated_set)


if __name__ == "__main__":
    unittest.main()



"""Gaussian functions for visualization.

Much of the visualization code in this file originates from projects of Christian W. Feldmann:
    https://github.com/c-feldmann/rdkit_heatmaps
    https://github.com/c-feldmann/compchemkit
"""

import numpy as np
import numpy.typing as npt


class GaussFunctor2D:  # pylint: disable=too-few-public-methods
    """2D Gaussian functor."""

    def __init__(
        self,
        center: npt.NDArray[np.float64],
        std1: float = 1,
        std2: float = 1,
        scale: float = 1,
        rotation: float = 0,
    ) -> None:
        """Initialize 2D Gaussian functor.

        Parameters
        ----------
        center: npt.NDArray[np.float64]
            Center of the Gaussian function.
        std1: float
            Standard deviation along the first axis.
        std2: float
            Standard deviation along the second axis.
        scale: float
            Scaling factor.
        rotation: float
            Rotation angle in radians.
        """
        self.center = center
        self.std = np.array([std1, std2]) ** 2  # scale stds to variance
        self.scale = scale
        self.rotation = rotation

        self._a = np.cos(self.rotation) ** 2 / (2 * self.std[0]) + np.sin(
            self.rotation
        ) ** 2 / (2 * self.std[1])
        self._b = -np.sin(2 * self.rotation) / (4 * self.std[0]) + np.sin(
            2 * self.rotation
        ) / (4 * self.std[1])
        self._c = np.sin(self.rotation) ** 2 / (2 * self.std[0]) + np.cos(
            self.rotation
        ) ** 2 / (2 * self.std[1])

    def __call__(self, pos: npt.NDArray[np.float64]) -> npt.NDArray[np.float64]:
        """Evaluate the Gaussian function at the given positions.

        Parameters
        ----------
        pos: npt.NDArray[np.float64]
            Array of positions to evaluate the Gaussian function at.

        Returns
        -------
        npt.NDArray[np.float64]
            Array of function values at the given positions.
        """
        exponent = self._a * (pos[:, 0] - self.center[0]) ** 2
        exponent += (
            2 * self._b * (pos[:, 0] - self.center[0]) * (pos[:, 1] - self.center[1])
        )
        exponent += self._c * (pos[:, 1] - self.center[1]) ** 2
        return self.scale * np.exp(-exponent)



"""Test module for behavior of chemprop in a pipeline."""

import unittest
from io import BytesIO
from typing import TypeVar

import joblib
import numpy as np
import pandas as pd
from lightning import pytorch as pl
from sklearn.base import clone
from sklearn.calibration import CalibratedClassifierCV

from molpipeline.any2mol import SmilesToMol
from molpipeline.error_handling import ErrorFilter, FilterReinserter
from molpipeline.estimators.chemprop.abstract import ABCChemprop
from molpipeline.estimators.chemprop.component_wrapper import (
    MPNN,
    BinaryClassificationFFN,
    BondMessagePassing,
    SumAggregation,
)
from molpipeline.estimators.chemprop.models import (
    ChempropClassifier,
    ChempropModel,
    ChempropMulticlassClassifier,
    ChempropRegressor,
)
from molpipeline.mol2any import MolToSmiles
from molpipeline.mol2any.mol2chemprop import MolToChemprop
from molpipeline.pipeline import Pipeline
from molpipeline.post_prediction import PostPredictionWrapper
from test_extras.test_chemprop.chemprop_test_utils.compare_models import compare_params
from tests import TEST_DATA_DIR


def get_smiles_checker_pipeline() -> Pipeline:
    """Get a pipeline that reads and writes the SMILES string.

    Invalid SMILES strings are replaced with NaN.

    Returns
    -------
    Pipeline
        The pipeline that reads and writes the SMILES string.
    """
    smiles2mol = SmilesToMol()
    mol2smiles = MolToSmiles()
    error_filter = ErrorFilter(filter_everything=True)
    filter_reinserter = FilterReinserter.from_error_filter(
        error_filter, fill_value=np.nan
    )
    smiles_pipeline = Pipeline(
        [
            ("smiles2mol", smiles2mol),
            ("error_filter", error_filter),
            ("mol2smiles", mol2smiles),
            ("filter_reinserter", filter_reinserter),
        ],
    )
    return smiles_pipeline


# pylint: disable=duplicate-code
def get_model() -> ChempropModel:
    """Get the Chemprop model.

    Returns
    -------
    ChempropModel
        The Chemprop model.
    """
    binary_clf_ffn = BinaryClassificationFFN()
    aggregate = SumAggregation()
    bond_message_passing = BondMessagePassing()
    mpnn = MPNN(
        message_passing=bond_message_passing,
        agg=aggregate,
        predictor=binary_clf_ffn,
    )
    chemprop_model = ChempropModel(model=mpnn)
    return chemprop_model


def get_model_pipeline() -> Pipeline:
    """Get the Chemprop model pipeline.

    Returns
    -------
    Pipeline
        The Chemprop model pipeline.
    """
    chemprop_model = get_model()
    mol2chemprop = MolToChemprop()
    smiles2mol = SmilesToMol()
    model_pipeline = Pipeline(
        steps=[
            ("smiles2mol", smiles2mol),
            ("mol2chemprop", mol2chemprop),
            ("model", chemprop_model),
        ],
    )
    return model_pipeline


DEFAULT_TRAINER = pl.Trainer(
    accelerator="cpu",
    logger=False,
    enable_checkpointing=False,
    max_epochs=5,
    enable_model_summary=False,
    enable_progress_bar=False,
    val_check_interval=0.0,
)


def get_regression_pipeline() -> Pipeline:
    """Get the Chemprop model pipeline for regression.

    Returns
    -------
    Pipeline
        The Chemprop model pipeline for regression.
    """

    smiles2mol = SmilesToMol()
    mol2chemprop = MolToChemprop()
    error_filter = ErrorFilter(filter_everything=True)
    filter_reinserter = FilterReinserter.from_error_filter(
        error_filter, fill_value=np.nan
    )
    chemprop_model = ChempropRegressor(lightning_trainer=DEFAULT_TRAINER)
    model_pipeline = Pipeline(
        steps=[
            ("smiles2mol", smiles2mol),
            ("mol2chemprop", mol2chemprop),
            ("error_filter", error_filter),
            ("model", chemprop_model),
            ("filter_reinserter", PostPredictionWrapper(filter_reinserter)),
        ],
    )
    return model_pipeline


def get_classification_pipeline() -> Pipeline:
    """Get the Chemprop model pipeline for classification.

    Returns
    -------
    Pipeline
        The Chemprop model pipeline for classification.
    """
    smiles2mol = SmilesToMol()
    mol2chemprop = MolToChemprop()
    error_filter = ErrorFilter(filter_everything=True)
    filter_reinserter = FilterReinserter.from_error_filter(
        error_filter, fill_value=np.nan
    )
    chemprop_model = ChempropClassifier(lightning_trainer=DEFAULT_TRAINER)
    model_pipeline = Pipeline(
        steps=[
            ("smiles2mol", smiles2mol),
            ("mol2chemprop", mol2chemprop),
            ("error_filter", error_filter),
            ("model", chemprop_model),
            ("filter_reinserter", PostPredictionWrapper(filter_reinserter)),
        ],
    )
    return model_pipeline


def get_multiclass_classification_pipeline(n_classes: int) -> Pipeline:
    """Get the Chemprop model pipeline for multiclass classification.

    Parameters
    ----------
    n_classes : int
        The number of classes for model initialization.

    Returns
    -------
    Pipeline
        The Chemprop model pipeline for multiclass classification.
    """
    smiles2mol = SmilesToMol()
    mol2chemprop = MolToChemprop()
    error_filter = ErrorFilter(filter_everything=True)
    filter_reinserter = FilterReinserter.from_error_filter(
        error_filter, fill_value=np.nan
    )
    chemprop_model = ChempropMulticlassClassifier(
        n_classes=n_classes, lightning_trainer=DEFAULT_TRAINER
    )
    model_pipeline = Pipeline(
        steps=[
            ("smiles2mol", smiles2mol),
            ("mol2chemprop", mol2chemprop),
            ("error_filter", error_filter),
            ("model", chemprop_model),
            ("filter_reinserter", PostPredictionWrapper(filter_reinserter)),
        ],
    )
    return model_pipeline


_T = TypeVar("_T")


def joblib_dump_load(obj: _T) -> _T:
    """Dump and load an object using joblib.

    Notes
    -----
    The object is not dumped to disk but to a BytesIO object.

    Parameters
    ----------
    obj : _T
        The object to dump and load.

    Returns
    -------
    _T
        The loaded object.
    """
    bytes_container = BytesIO()
    joblib.dump(obj, bytes_container)
    bytes_container.seek(0)  # update to enable reading
    bytes_model = bytes_container.read()

    return joblib.load(BytesIO(bytes_model))


class TestChempropPipeline(unittest.TestCase):
    """Test the Chemprop model pipeline."""

    def test_get_set_params(self) -> None:
        """Test the get_params and set_params methods of the pipeline."""
        model = get_model_pipeline()
        orig_params = model.get_params(deep=True)
        new_params = {
            "model__model__predictor__activation": "relu",
            "model__model__predictor__dropout": 0.5,
            "model__model__message_passing__activation": "relu",
        }
        # Check setting new parameters
        model.set_params(**new_params)
        model_params = model.get_params(deep=True)
        for param_name, param in new_params.items():
            self.assertEqual(param, model_params[param_name])
        model.set_params(**orig_params)
        model_params = model.get_params(deep=True)
        for param_name, param in orig_params.items():
            self.assertEqual(param, model_params[param_name])

    def test_clone(self) -> None:
        """Test the clone method of the pipeline."""
        model = get_model_pipeline()
        cloned_model = clone(model)
        for step_name, step in model.steps:
            cloned_step = cloned_model.named_steps[step_name]
            self.assertEqual(step.__class__, cloned_step.__class__)
            params = step.get_params(deep=True)  # type: ignore
            cloned_params = cloned_step.get_params(deep=True)
            if isinstance(step, ABCChemprop):
                compare_params(self, step, cloned_step)
                continue
            for param_name, param in params.items():
                # If parm implements get_params, it was cloned as well and we need
                # to compare the parameters. Since all parameters are listed flat in
                # the params dicts, all objects are identical if param dicts are identical.
                if hasattr(param, "get_params"):
                    self.assertEqual(
                        param.__class__, cloned_params[param_name].__class__
                    )
                else:
                    self.assertEqual(
                        param, cloned_params[param_name], f"Failed for {param_name}"
                    )

    def test_passing_smiles(self) -> None:
        """Test passing SMILES strings to the pipeline.

        Since the pipeline is not trained, we only check if the prediction is successful.
        """
        model = get_model_pipeline()
        smiles = ["CCO", "CNC", "CCN"]
        pred = model.predict(smiles)
        self.assertEqual(len(pred), 3)
        proba = model.predict_proba(smiles)
        self.assertEqual(len(proba), 3)

    def test_error_handling(self) -> None:
        """Test the error handling in the pipeline."""
        error_filter = ErrorFilter(filter_everything=True)
        filter_reinserter = FilterReinserter.from_error_filter(
            error_filter, fill_value=np.nan
        )
        pipeline = Pipeline(
            steps=[
                ("smiles2mol", SmilesToMol()),
                ("mol2chemprop", MolToChemprop()),
                ("error_filter", error_filter),
                ("model", get_model()),
                ("filter_reinserter", PostPredictionWrapper(filter_reinserter)),
            ]
        )
        smiles = ["CCO", "CNC", "CCN", "invalid"]
        pred = pipeline.predict(smiles)
        self.assertEqual(len(pred), 4)
        self.assertTrue(np.isnan(pred[-1]))
        proba = pipeline.predict_proba(smiles)
        self.assertEqual(len(proba), 4)
        self.assertTrue(np.isnan(proba[-1]).all())


class TestRegressionPipeline(unittest.TestCase):
    """Test the Chemprop model pipeline for regression."""

    def test_prediction(self) -> None:
        """Test the prediction of the regression model."""

        molecule_net_logd_df = pd.read_csv(
            TEST_DATA_DIR / "molecule_net_logd.tsv.gz", sep="\t", nrows=100
        )
        regression_model = get_regression_pipeline()
        regression_model.fit(
            molecule_net_logd_df["smiles"].tolist(),
            molecule_net_logd_df["exp"].to_numpy(),
        )
        pred = regression_model.predict(molecule_net_logd_df["smiles"].tolist())

        self.assertEqual(len(pred), len(molecule_net_logd_df))

        model_copy = joblib_dump_load(regression_model)
        pred_copy = model_copy.predict(molecule_net_logd_df["smiles"].tolist())
        self.assertTrue(np.allclose(pred, pred_copy))

        # Test single prediction, this was causing an error before
        single_mol_pred = regression_model.predict(
            [molecule_net_logd_df["smiles"].iloc[0]]
        )
        self.assertEqual(single_mol_pred.shape, (1,))


class TestClassificationPipeline(unittest.TestCase):
    """Test the Chemprop model pipeline for classification."""

    def setUp(self) -> None:
        """Set up repeated variables."""
        molecule_net_bbbp_df = pd.read_csv(
            TEST_DATA_DIR / "molecule_net_bbbp.tsv.gz", sep="\t", nrows=100
        )
        smiles_pipeline = get_smiles_checker_pipeline()
        molecule_net_bbbp_df["smiles"] = smiles_pipeline.transform(
            molecule_net_bbbp_df["smiles"]
        )
        molecule_net_bbbp_df = molecule_net_bbbp_df.dropna(subset=["smiles", "p_np"])

        self.molecule_net_bbbp_df = molecule_net_bbbp_df

    def test_prediction(self) -> None:
        """Test the prediction of the classification model."""
        classification_model = get_classification_pipeline()
        classification_model.fit(
            self.molecule_net_bbbp_df["smiles"].tolist(),
            self.molecule_net_bbbp_df["p_np"].to_numpy(),
        )
        pred = classification_model.predict(
            self.molecule_net_bbbp_df["smiles"].tolist()
        )
        proba = classification_model.predict_proba(
            self.molecule_net_bbbp_df["smiles"].tolist()
        )
        self.assertEqual(len(pred), len(self.molecule_net_bbbp_df))
        self.assertEqual(proba.shape[1], 2)
        self.assertEqual(proba.shape[0], len(self.molecule_net_bbbp_df))

        model_copy = joblib_dump_load(classification_model)
        pred_copy = model_copy.predict(self.molecule_net_bbbp_df["smiles"].tolist())
        proba_copy = model_copy.predict_proba(
            self.molecule_net_bbbp_df["smiles"].tolist()
        )

        nan_indices = np.isnan(pred)
        self.assertListEqual(nan_indices.tolist(), np.isnan(pred_copy).tolist())
        self.assertTrue(np.allclose(pred[~nan_indices], pred_copy[~nan_indices]))

        self.assertEqual(proba.shape, proba_copy.shape)
        self.assertTrue(np.allclose(proba[~nan_indices], proba_copy[~nan_indices]))

        # Test single prediction, this was causing an error before
        single_mol_pred = classification_model.predict(
            [self.molecule_net_bbbp_df["smiles"].iloc[0]]
        )
        self.assertEqual(single_mol_pred.shape, (1,))
        single_mol_proba = classification_model.predict_proba(
            [self.molecule_net_bbbp_df["smiles"].iloc[0]]
        )
        self.assertEqual(single_mol_proba.shape, (1, 2))

    def test_calibrated_classifier(self) -> None:
        """Test if the pipeline can be used with a CalibratedClassifierCV."""
        calibrated_pipeline = CalibratedClassifierCV(
            get_classification_pipeline(), cv=2, ensemble=True, method="isotonic"
        )
        calibrated_pipeline.fit(
            self.molecule_net_bbbp_df["smiles"].tolist(),
            self.molecule_net_bbbp_df["p_np"].to_numpy(),
        )
        predicted_value_array = calibrated_pipeline.predict(
            self.molecule_net_bbbp_df["smiles"].tolist()
        )
        predicted_proba_array = calibrated_pipeline.predict_proba(
            self.molecule_net_bbbp_df["smiles"].tolist()
        )
        self.assertIsInstance(predicted_value_array, np.ndarray)
        self.assertIsInstance(predicted_proba_array, np.ndarray)
        self.assertEqual(
            predicted_value_array.shape,
            (len(self.molecule_net_bbbp_df["smiles"].tolist()),),
        )
        self.assertEqual(
            predicted_proba_array.shape,
            (len(self.molecule_net_bbbp_df["smiles"].tolist()), 2),
        )


class TestMulticlassClassificationPipeline(unittest.TestCase):
    """Test the Chemprop model pipeline for multiclass classification."""

    def test_prediction(self) -> None:
        """Test the prediction of the multiclass classification model."""

        test_data_df = pd.read_csv(
            TEST_DATA_DIR / "multiclass_mock.tsv", sep="\t", index_col=False
        )
        classification_model = get_multiclass_classification_pipeline(n_classes=3)
        mols = test_data_df["Molecule"].tolist()
        classification_model.fit(
            mols,
            test_data_df["Label"].to_numpy(),
        )
        pred = classification_model.predict(mols)
        proba = classification_model.predict_proba(mols)
        self.assertEqual(len(pred), len(test_data_df))
        self.assertEqual(proba.shape[1], 3)
        self.assertEqual(proba.shape[0], len(test_data_df))

        model_copy = joblib_dump_load(classification_model)
        pred_copy = model_copy.predict(mols)
        proba_copy = model_copy.predict_proba(mols)

        nan_mask = np.isnan(pred)
        self.assertListEqual(nan_mask.tolist(), np.isnan(pred_copy).tolist())
        self.assertTrue(np.allclose(pred[~nan_mask], pred_copy[~nan_mask]))

        self.assertEqual(proba.shape, proba_copy.shape)
        self.assertEqual(pred.shape, pred_copy.shape)
        self.assertTrue(np.allclose(proba[~nan_mask], proba_copy[~nan_mask]))

        # Test single prediction, this was causing an error before
        single_mol_pred = classification_model.predict(
            [test_data_df["Molecule"].iloc[0]]
        )
        self.assertEqual(single_mol_pred.shape, (1,))
        single_mol_proba = classification_model.predict_proba(
            [test_data_df["Molecule"].iloc[0]]
        )
        self.assertEqual(single_mol_proba.shape, (1, 3))

        with self.assertRaises(ValueError):
            classification_model.fit(
                mols,
                test_data_df["Label"].add(1).to_numpy(),
            )
        with self.assertRaises(ValueError):
            classification_model = get_multiclass_classification_pipeline(n_classes=2)
            classification_model.fit(
                mols,
                test_data_df["Label"].to_numpy(),
            )
